# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrappers for the json returned by the ThreatExchange API to typed objects.
"""

import typing as t


class ThreatDescriptor(t.NamedTuple):
    """
    Wrapper around ThreatExchange JSON for a ThreatDescriptor.

    Example:
    {
        "id": "3058061737574159"
        "raw_indicator": "facefacefacefacefacefacefaceface",
        "type": "HASH_MD5",
        "owner": {
            "id": "616912251743987",
            ...,
        },
        "status": "MALICIOUS",
        "tags": null,
        "added_on": "2020-07-01T18:31:15+0000"
        ...
    }
    """

    # TODO - do something smarter than this - static
    #        class variable problematic, currently set in main.py
    MY_APP_ID = -1  # type: ignore

    # You declared the indicator was in the collaboration label set
    TRUE_POSITIVE = "true_positive"  # type: ignore
    # You declared the indicator was not in the collaboration label set
    FALSE_POSITIVE = "false_positive"  # type: ignore
    # Someone declared the indicator was not in the collaboration label set
    DISPUTED = "disputed"  # type: ignore

    # Special tags to mark whether you (or someone else)
    # has weighed in on the indicator
    SPECIAL_TAGS = frozenset((TRUE_POSITIVE, FALSE_POSITIVE, DISPUTED))  # type: ignore

    id: int
    raw_indicator: str
    indicator_type: str
    owner_id: int
    tags: t.List[str]
    status: str
    added_on: str

    @classmethod
    def from_te_json(cls, my_app_id: t.Union[str, int], td_json) -> "ThreatDescriptor":
        # Hack for now, but nearly refactored out of cls state
        cls.MY_APP_ID = my_app_id
        owner_id_str = td_json["owner"]["id"]
        tags = td_json.get("tags", [])
        # This is needed because ThreatExchangeAPI.get_threat_descriptors()
        # does a transform, but other locations do not
        if isinstance(tags, dict):
            tags = sorted(tag["text"] for tag in tags["data"])
        td = cls(  # type: ignore
            id=int(td_json["id"]),
            raw_indicator=td_json["raw_indicator"],
            indicator_type=td_json["type"],
            owner_id=int(owner_id_str),
            tags=[tag for tag in tags if tag not in ThreatDescriptor.SPECIAL_TAGS],
            status=td_json["status"],
            added_on=td_json["added_on"],
        )
        # Add special tags
        # TODO - Consider stripping out collab labels
        #        from FALSE_POSITIVE & NON_MALICIOUS
        # Is this my descriptor?
        if td.is_mine:
            if td.status == "NON_MALICIOUS":
                td.tags.append(ThreatDescriptor.FALSE_POSITIVE)
            else:
                td.tags.append(ThreatDescriptor.TRUE_POSITIVE)
        # Disputed path #1 - mark as non_malicious
        elif td.status == "NON_MALICIOUS":
            td.tags.append(ThreatDescriptor.DISPUTED)
        # Disputed path #2 - react with DISAGREE_WITH_TAGS
        elif "DISAGREE_WITH_TAGS" in td_json.get("my_reactions", ()):
            td.tags.append(ThreatDescriptor.FALSE_POSITIVE)
        elif any(
            t == "DISAGREE_WITH_TAGS" for r in td_json.get("reactions", []) for t in r
        ):
            td.tags.append(ThreatDescriptor.DISPUTED)
        return td

    def to_params(self) -> t.Dict[str, t.Any]:
        params = dict(self.__dict__)
        params["type"] = params.pop("indicator_type")
        if not params["tags"]:
            del params["tags"]
        return params

    @property
    def is_true_positive(self) -> bool:
        return self.TRUE_POSITIVE in self.tags

    @property
    def is_false_positive(self) -> bool:
        return self.FALSE_POSITIVE in self.tags

    @property
    def is_mine(self) -> bool:
        """This Descriptor is my App's Opinion"""
        return self.MY_APP_ID == self.owner_id


class SimpleDescriptorRollup:
    """
    A simple way to merge opinions on the same indicator.

    This contains all the information needed for simple SignalType state.
    """

    IS_MY_OPINION = "mine"

    __slots__ = ["first_descriptor_id", "added_on", "labels"]

    def __init__(
        self, first_descriptor_id: int, added_on: str, labels: t.Set[str]
    ) -> None:
        self.first_descriptor_id = first_descriptor_id
        self.added_on = added_on  # TODO - convert to int?
        self.labels = set(labels)

    @classmethod
    def from_descriptor(cls, descriptor: ThreatDescriptor) -> "SimpleDescriptorRollup":
        return cls(descriptor.id, descriptor.added_on, set(descriptor.tags))

    @classmethod
    def from_descriptors(
        cls, descriptors: t.Iterable[ThreatDescriptor]
    ) -> "SimpleDescriptorRollup":
        ret = None
        for d in descriptors:
            if not ret:
                ret = cls.from_descriptor(d)
            else:
                ret.merge(d)
        if ret:
            return ret
        raise ValueError("Empty descriptor list!")

    def merge(self, descriptor: ThreatDescriptor) -> None:
        # Is the other descriptor mine? If so, unconditionally take it and clear
        # everything else
        if descriptor.is_mine:
            self.added_on = self.IS_MY_OPINION
            self.first_descriptor_id = descriptor.id
            self.labels = set(descriptor.tags)
            return
        # My descriptor beats my reactions, and I don't want
        # to take anyone else's opinion
        elif self.added_on == self.IS_MY_OPINION:
            return
        # Is my reaction?
        elif descriptor.is_false_positive:
            self.added_on = self.IS_MY_OPINION
            self.first_descriptor_id = descriptor.id
            self.labels = set(descriptor.tags)
            return
        # Else merge the labels together
        self.added_on, self.first_descriptor_id = min(
            (self.added_on, self.first_descriptor_id),
            (descriptor.added_on, descriptor.id),
        )
        self.labels.union(descriptor.tags)

    def as_row(self) -> t.Tuple[int, str, str]:
        """Simple conversion to CSV row"""
        return self.first_descriptor_id, self.added_on, " ".join(self.labels)

    @classmethod
    def from_row(cls, row: t.List[str]) -> "SimpleDescriptorRollup":
        """Simple conversion from CSV row"""
        labels = []
        if row[2]:
            labels = row[2].split(" ")
        return cls(int(row[0]), row[1], set(labels))

    @classmethod
    def from_threat_updates_json(
        cls, my_app_id: int, te_json: t.Dict[str, t.Any]
    ) -> t.Optional["SimpleDescriptorRollup"]:
        if te_json["should_delete"]:
            return None
        # https://github.com/facebook/ThreatExchange/issues/834
        if not te_json.get("descriptors", {}).get("data"):
            return None
        descriptors = []
        for descriptor_json in te_json["descriptors"]["data"]:
            # Look at me ma! I'm modifying input paramaters!
            descriptor_json["raw_indicator"] = te_json["indicator"]
            descriptor_json["type"] = te_json["type"]
            descriptors.append(
                ThreatDescriptor.from_te_json(my_app_id, descriptor_json)
            )
        return cls.from_descriptors(descriptors)

    @staticmethod
    def te_threat_updates_fields() -> t.Tuple[str, ...]:
        return (
            "id",
            "indicator",
            "type",
            "last_updated",
            "should_delete",
            "descriptors{%s}"
            % ",".join(
                (
                    "reactions",
                    "my_reactions",
                    "owner{id}",
                    "tags",
                    "status",
                    "added_on",
                )
            ),
        )



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the pdf signal type.
"""

import re
import typing as t
from threatexchange.content_type.content_base import ContentType
from threatexchange.content_type.text import TextContent

from threatexchange.signal_type import signal_base
from threatexchange.signal_type.raw_text import RawTextSignal

import tlsh

TLSH_CONFIDENT_MATCH_THRESHOLD = 30
EXPECT_TLSH_HASH_LENGTH = 72


class TextTLSHSignal(signal_base.SimpleSignalType, signal_base.TextHasher):
    """
    Simple signal type for text using TLSH.

    Read about TLSH at https://github.com/trendmicro/tlsh
    """

    INDICATOR_TYPE = "HASH_TEXT_TLSH"

    @classmethod
    def get_content_types(cls) -> t.List[t.Type[ContentType]]:
        return [TextContent]

    @classmethod
    def validate_signal_str(cls, signal_str: str) -> str:
        """'T1' followed 70 hexidecimal characters. Total length 72 characters."""
        if not re.match("^T1[0-9A-F]{70}$", signal_str):
            raise ValueError("invalid TLSH hash")
        return signal_str

    @classmethod
    def hash_from_str(cls, text: str) -> str:
        hash_str = str(tlsh.hash(text.encode()))
        if hash_str == "TNULL":  # Likely too short
            return ""
        return hash_str

    @classmethod
    def compare_hash(
        cls,
        hash1: str,
        hash2: str,
        tlsh_threshold: int = TLSH_CONFIDENT_MATCH_THRESHOLD,
    ) -> signal_base.SignalComparisonResult:
        dist: int = tlsh.diffxlen(hash1, hash2)
        return signal_base.SignalComparisonResult.from_simple_dist(dist, tlsh_threshold)

    @staticmethod
    def get_examples() -> t.List[str]:
        return [TextTLSHSignal.hash_from_str(s) for s in RawTextSignal.get_examples()]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved


from io import StringIO
import pathlib
import shutil
import sys
import tempfile
import typing as t
from unittest.mock import patch
import unittest
import pytest

from threatexchange.cli.main import inner_main
from threatexchange.cli.exceptions import CommandError


class E2ETestSystemExit(Exception):
    def __init__(self, code: int) -> None:
        super().__init__()
        self.returncode = code


class ThreatExchangeCLIE2eHelper:

    # A keystroke saver, prefix these args to the beginning
    # of every call.
    # You can bypass by making the first arg "tx" or
    # "threatexchange".
    COMMON_CALL_ARGS: t.Sequence[str] = ()

    _state_dir: pathlib.Path

    def cli_call(self, *given_args: str) -> str:
        """
        Call the threatexchange CLI
        """
        args: t.List[str] = []
        if next(iter(given_args), None) in ("tx", "threatexchange"):
            given_args = given_args[1:]
        else:
            args.extend(self.COMMON_CALL_ARGS)
        args.extend(given_args)
        print("Calling: $ threatexchange", " ".join(args), file=sys.stderr)
        with patch("sys.stdout", new=StringIO(newline=None)) as fake_out, patch(
            "sys.argv", new=["threatexchange"] + args
        ):
            try:
                inner_main(args, state_dir=self._state_dir)
            except SystemExit as se:
                if se.code != 0:
                    raise E2ETestSystemExit(se.code)
            return fake_out.getvalue()

    def assert_cli_output(
        self,
        args: t.Iterable[str],
        expected_output: t.Union[str, t.Iterable[str], t.Dict[int, str]],
    ) -> None:
        output = self.cli_call(*args)
        if isinstance(expected_output, str):
            assert expected_output.strip() == output.strip()
            return
        lines = output.strip().split("\n")
        if not isinstance(expected_output, dict):
            expected_output = dict(enumerate(expected_output))
        for line, expected_line_output in expected_output.items():
            if line < 0:
                assert line >= -len(lines)
            else:
                assert line < len(lines)
            assert lines[line] == expected_line_output

    def assert_cli_usage_error(
        self, args: t.Iterable[str], msg_regex: str = None
    ) -> None:
        with pytest.raises((CommandError, E2ETestSystemExit), match=msg_regex) as ex:
            self.cli_call(*args)
        exception = t.cast(t.Union[CommandError, E2ETestSystemExit], ex.value)
        assert exception.returncode == 2


@pytest.fixture
def te_cli(tmp_path: pathlib.Path) -> ThreatExchangeCLIE2eHelper:
    ret = ThreatExchangeCLIE2eHelper()
    ret._state_dir = tmp_path
    return ret


class ThreatExchangeCLIE2eTest(unittest.TestCase, ThreatExchangeCLIE2eHelper):
    def setUp(self) -> None:
        state_dir = pathlib.Path(tempfile.mkdtemp())
        self.addCleanup(lambda: shutil.rmtree(str(state_dir)))
        self._state_dir = state_dir



from threatexchange.cli.tests.e2e_test_helper import ThreatExchangeCLIE2eTest

from threatexchange.cli.main import _DEFAULT_SIGNAL_TYPES


class DatasetCommandTest(ThreatExchangeCLIE2eTest):
    def test(self):
        """
        Test on the output of the dataset command, using the sample signals

        The sample signals will probably change a few times during the course
        of development, which will unexpectedly break this test, so apologies
        future developers.
        """
        signal_count = sum(len(st.get_examples()) for st in _DEFAULT_SIGNAL_TYPES)

        # Since we didn't set a collab config, this will fetch on first call
        self.assert_cli_output(
            ("dataset",),
            [
                "pdq: 16",
                "raw_text: 3",
                "video_md5: 2",
                "trend_query: 1",
                "url: 1",
                "url_md5: 1",
            ],
        )
        self.assert_cli_output(
            ("dataset", "--signal-summary"),
            [
                "pdq: 16",
                "raw_text: 3",
                "video_md5: 2",
                "trend_query: 1",
                "url: 1",
                "url_md5: 1",
            ],
        )

        # The sort of printed output is currently not stable
        output = self.cli_call("dataset", "-P")
        assert output.count("\n") == signal_count
        assert (
            "url "
            "https://developers.facebook.com/docs/threat-exchange/reference/apis/ "
            "'Sample Signals' "
            "INVESTIGATION_SEED"
        ) in output
        # The filters change the print output
        self.assert_cli_output(
            ("dataset", "-P", "-s", "url"),
            "https://developers.facebook.com/docs/threat-exchange/reference/apis/ "
            "'Sample Signals' "
            "INVESTIGATION_SEED",
        )
        self.assert_cli_output(
            ("dataset", "-P", "-s", "url", "-c", "Sample Signals"),
            "https://developers.facebook.com/docs/threat-exchange/reference/apis/ "
            "INVESTIGATION_SEED",
        )
        self.assert_cli_output(
            ("dataset", "-P", "-s", "url", "-S"),
            "https://developers.facebook.com/docs/threat-exchange/reference/apis/",
        )
        # CSV stuff
        output = self.cli_call("dataset", "-P", "--csv")
        assert output.count("\n") - 1 == signal_count  # -1 for header
        assert (
            "url,"
            "https://developers.facebook.com/docs/threat-exchange/reference/apis/,"
            "Sample Signals,"
            "INVESTIGATION_SEED,"
        ) in output
        # Repeat same filters - however, these don't change the output format except -S\
        csv_header = "signal_type,signal_str,collab,category,tags\n"
        self.assert_cli_output(
            ("dataset", "-P", "--csv", "-s", "url"),
            csv_header + "url,"
            "https://developers.facebook.com/docs/threat-exchange/reference/apis/,"
            "Sample Signals,INVESTIGATION_SEED,",
        )
        self.assert_cli_output(
            ("dataset", "-P", "--csv", "-s", "url", "-c", "Sample Signals"),
            csv_header + "url,"
            "https://developers.facebook.com/docs/threat-exchange/reference/apis/,"
            "Sample Signals,INVESTIGATION_SEED,",
        )
        # --csv and -S not combinable
        self.assert_cli_usage_error(("dataset", "-P", "--csv", "-S"))

    def test_indices(self):
        self.cli_call("fetch", "--skip-index-rebuild")
        self.cli_call("dataset", "-r")  # Someday actually test?



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import pytest
import typing as t

from threatexchange.signal_type.signal_base import (
    FileHasher,
    MatchesStr,
    SignalType,
    TextHasher,
)


class SignalTypeHashTest:
    """
    Helpers to help you quickly test common functionality

    Most functions are designed to be used to be used in a provider fasion
    to rapidly make simple tests.
    """

    TYPE: t.ClassVar[t.Type[SignalType]]

    def assert_signal_str_valid(
        self, s: str, expected: t.Optional[str] = None, *, valid: bool = True
    ):
        if not valid:
            self.assert_signal_str_invalid(s)
            return
        expected = s if expected is None else expected
        validated = self.TYPE.validate_signal_str(s)
        assert expected == validated

    def assert_signal_str_invalid(
        self, s: str, expected_exception=Exception, *args, **kwargs
    ) -> None:
        with pytest.raises(expected_exception, *args, **kwargs):
            self.TYPE.validate_signal_str(s)


THashValidateCase = t.Union[str, t.Tuple[str, t.Union[None, str, t.Type[Exception]]]]
THashCompareCase = t.Union[
    t.Tuple[str, str], t.Tuple[str, str, bool], t.Tuple[str, str, bool, t.Any]
]


class SignalTypeAutoTest(SignalTypeHashTest):
    """
    A helper to automatically generate a bunch of simple tests for a SignalType.

    If you create a test class that extends this one, it will automatically create
    test cases, which you can provide test data by implementing the *_cases().

    check out test_raw_text.py in this directory for a simple example

    This is probably not the cleanest way you could have done this, but it's
    probably better than nothing.
    """

    def test_get_name(self):
        assert self.TYPE.get_name()
        # Could also test for collisions, but ¯\_(ツ)_/¯

    def test_examples(self):
        examples = self.TYPE.get_examples()
        # It's possible that there some SignalTypes where this doesn't make
        # sense, in which case you can override this test to just always pass
        assert examples, f"Add some examples to {self.TYPE.__name__}.get_examples()"
        for example in examples:
            self.assert_signal_str_valid(example)
            if issubclass(self.TYPE, FileHasher):
                assert self.TYPE.compare_hash(
                    example, example
                ).match, f"Case: {example}"

    def get_validate_hash_cases(self) -> t.Iterable[THashValidateCase]:
        raise NotImplementedError

    def test_validate_hash(self):
        for t in self.get_validate_hash_cases():
            if isinstance(t, str):
                t = (t, None)
            s, expected = t
            if isinstance(expected, type):  # Exception
                self.assert_signal_str_invalid(s, expected)
            else:
                self.assert_signal_str_valid(s, expected)

    def get_compare_hash_cases(self) -> t.Iterable[THashValidateCase]:
        raise NotImplementedError

    def test_compare_hash(self):
        for t in self.get_compare_hash_cases():
            assert issubclass(self.TYPE, FileHasher)  # For typing
            a, b = t[:2]
            expect_match = t[2] if len(t) > 2 else True

            result = self.TYPE.compare_hash(a, b)
            assert result.match == expect_match, f"Case: {t}"
            if len(t) > 3:
                assert result.distance == t[3]


class TextHasherAutoTest(SignalTypeAutoTest):
    """
    Auto-testing, but with TextHasher classes
    """

    def get_hashes_from_str_cases(self) -> t.Iterable[t.Tuple[str, str]]:
        raise NotImplementedError

    def test_hash_from_str(self):
        assert issubclass(self.TYPE, TextHasher)  # For typing
        for s, hashed in self.get_hashes_from_str_cases():
            assert self.TYPE.hash_from_str(s) == hashed, f"Case: {(s, hashed)}"


TMatchesStrCase = t.Union[
    t.Tuple[str, str], t.Tuple[str, str, bool], t.Tuple[str, str, bool, t.Any]
]


class MatchesStrAutoTest(SignalTypeAutoTest):
    """
    Auto-testing, but with MatchesStr classes
    """

    def get_matches_str_cases(self) -> t.Iterable[TMatchesStrCase]:
        raise NotImplementedError

    def test_matches_str(self):
        for t in self.get_matches_str_cases():
            signal, haystack = t[:2]
            expect_match = t[2] if len(t) > 2 else True

            result = self.TYPE.matches_str(signal, haystack)
            assert result.match == expect_match, f"Case: {t}"
            if len(t) > 3:
                assert result.distance.distance == t[3], f"Case: {t}"



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from threatexchange.extensions.manifest import ThreatExchangeExtensionManifest
from threatexchange.extensions.pdf.content import PDFContent


TX_MANIFEST = ThreatExchangeExtensionManifest(
    content_types=(PDFContent,),
)



#  Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Helpers for handling authentification and credentials for SignalExchangeAPI
"""

from abc import ABCMeta, abstractmethod
import contextlib
import logging
import os
import pathlib
import typing as t

from threatexchange.exchanges.signal_exchange_api import (
    AnySignalExchangeAPI,
    TCollabConfig,
)


class SignalExchangeAPIInvalidAuthException(Exception):
    """
    An exception you can use to hint users their authentification is bad

    This can be because it's incorrectly formatted, it's been expired,
    or a multitude of other reasons.
    """

    def __init__(self, src_api: t.Type[AnySignalExchangeAPI], message: str) -> None:
        self.src_api = src_api
        self.message = message


class SignalExchangeAPIMissingAuthException(Exception):
    """
    An exception you can use to hint users how to authentificate your API
    """

    def __init__(
        self,
        src_api: t.Type[AnySignalExchangeAPI],
        *,
        file_hint: str = "",
        env_hint: str = "",
    ) -> None:
        self.src_api = src_api
        self.hints: t.List[str] = []
        if env_hint:
            self.add_env_hint(env_hint)
        if file_hint:
            self.add_file_hint(file_hint)

    def add_file_hint(self, filename: str) -> None:
        self.hints.append(
            f"creating (and maybe `chmod 400`) a file with credentials at {filename}"
        )

    def add_env_hint(self, variable_name: str) -> None:
        self.hints.append(f"populating an environment variable called {variable_name}")

    def pretty_str(self) -> str:
        lines = [
            f"Couldn't authenticate {self.src_api.get_name()}, "
            "it's missing authentification!"
        ]
        if self.hints:
            lines.append("You can fix this by:")
            for i, hint in enumerate(self.hints, 1):
                lines.append(f"  {i}. {hint}")
        return "\n".join(lines)


CredentialSelf = t.TypeVar("CredentialSelf", bound="CredentialHelper")


class CredentialHelper:
    """
    Wrapper to help standardize credential sources, and tie to exceptions
    """

    ENV_VARIABLE: t.ClassVar[str] = ""
    FILE_NAME: t.ClassVar[str] = ""

    # This is slated to be removed in a future version!
    # It's a placeholder approach while for_collab() gains the auth argument
    _DEFAULT: t.ClassVar[t.Optional["CredentialHelper"]] = None  # t.Self
    _DEFAULT_SRC: t.ClassVar[str] = ""

    @classmethod
    def get(
        cls: t.Type[CredentialSelf], for_cls: t.Type[AnySignalExchangeAPI]
    ) -> CredentialSelf:
        srcs: t.List[t.Tuple[t.Callable[[], t.Optional[CredentialSelf]], str]] = [
            ((lambda: cls._DEFAULT), cls._DEFAULT_SRC),  # type: ignore[list-item,return-value]
            (cls._from_env, f"environment variable {cls.ENV_VARIABLE}"),
            (cls._from_file, f"file {cls.FILE_NAME}"),
        ]
        for fn, src in srcs:
            try:
                creds = fn()
                if creds is None:
                    continue
                if creds._are_valid():
                    return creds
            except Exception:
                logging.exception("Exception during parsing %s", src)
                pass
            raise SignalExchangeAPIInvalidAuthException(
                for_cls, f"Invalid credentials from {src}"
            )
        ex = SignalExchangeAPIMissingAuthException(for_cls)
        if cls._DEFAULT_SRC:
            ex.hints.append(cls._DEFAULT_SRC)
        if cls.ENV_VARIABLE:
            ex.add_env_hint(cls.ENV_VARIABLE)
        if cls.FILE_NAME:
            ex.add_file_hint(cls.FILE_NAME)
        raise ex

    @classmethod
    def set_default(
        cls: t.Type[CredentialSelf], new_default: t.Optional[CredentialSelf], src: str
    ) -> t.ContextManager[None]:
        """
        Set the default (highest preferred) credentials manually.

        They won't be checked for validity until a future get()

        This call can be used as contextmanager to unset within a `with` statement
        """
        cls._DEFAULT = new_default
        cls._DEFAULT_SRC = src
        return cls._unset_default_context()

    @classmethod
    @contextlib.contextmanager
    def _unset_default_context(cls: t.Type[CredentialSelf]) -> t.Iterator[None]:
        yield
        cls.clear_default()

    @classmethod
    def clear_default(cls) -> None:
        """Reset the default"""
        cls.set_default(None, "")

    @classmethod
    def _from_str(cls: t.Type[CredentialSelf], s: str) -> t.Optional[CredentialSelf]:
        """Parse credentials from a string"""
        return None

    @classmethod
    def _from_file(cls: t.Type[CredentialSelf]) -> t.Optional[CredentialSelf]:
        """Parse credentials from a file"""
        if not cls.FILE_NAME:
            return None
        path = pathlib.Path(cls.FILE_NAME).expanduser()
        if not path.is_file():
            return None
        return cls._from_str(path.read_text().strip())

    @classmethod
    def _from_env(cls: t.Type[CredentialSelf]) -> t.Optional[CredentialSelf]:
        """Parse credentials from an environment variable"""
        if not cls.ENV_VARIABLE:
            return None
        s = os.environ.get(cls.ENV_VARIABLE)
        if not s:
            return None
        return cls._from_str(s)

    def _are_valid(self) -> bool:
        return True


TCredentials = t.TypeVar("TCredentials", bound=CredentialHelper)
Self = t.TypeVar("Self")


class SignalExchangeWithAuth(t.Generic[TCollabConfig, TCredentials], metaclass=ABCMeta):
    """
    A mixin for SignalExchange APIs that need authentification/credentials

    The promises made by for_collab are the same - if no credentials are passed in,
    the class can search for its own credentials (though the CredentialHelper class
    makes it easy to do there).
    """

    @staticmethod
    @abstractmethod
    def get_credential_cls() -> t.Type[TCredentials]:
        pass

    @classmethod
    @abstractmethod
    def for_collab(
        cls: t.Type[Self],
        collab: TCollabConfig,
        credentials: t.Optional[TCredentials] = None,
    ) -> Self:
        """
        @see SignalExchangeAPI.for_collab

        If credentials are passed in, the API should use those rather than
        trying to discover its own.
        """
        pass



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Implementation of SignalTypeIndex abstraction for PDQ by wrapping
hashing.pdq_faiss_matcher.
"""

import typing as t

from threatexchange.signal_type.index import (
    IndexMatchUntyped,
    SignalSimilarityInfoWithIntDistance,
    SignalTypeIndex,
    T as IndexT,
)
from threatexchange.signal_type.pdq.pdq_faiss_matcher import (
    PDQMultiHashIndex,
    PDQFlatHashIndex,
    PDQHashIndex,
)

PDQIndexMatch = IndexMatchUntyped[SignalSimilarityInfoWithIntDistance, IndexT]


class PDQIndex(SignalTypeIndex[IndexT]):
    """
    Wrapper around the pdq faiss index lib using PDQMultiHashIndex
    """

    @classmethod
    def get_match_threshold(cls):
        return 31  # PDQ_CONFIDENT_MATCH_THRESHOLD

    @classmethod
    def _get_empty_index(cls) -> PDQHashIndex:
        return PDQMultiHashIndex()

    def __init__(self, entries: t.Iterable[t.Tuple[str, IndexT]] = ()) -> None:
        super().__init__()
        self.local_id_to_entry: t.List[t.Tuple[str, IndexT]] = []
        self.index: PDQHashIndex = self._get_empty_index()
        self.add_all(entries=entries)

    def __len__(self) -> int:
        return len(self.local_id_to_entry)

    def query(self, hash: str) -> t.Sequence[PDQIndexMatch[IndexT]]:
        """
        Look up entries against the index, up to the max supported distance.
        """

        # query takes a signal hash but index supports batch queries hence [hash]
        results = self.index.search_with_distance_in_result(
            [hash], self.get_match_threshold()
        )

        matches = []
        for id, _, distance in results[hash]:
            matches.append(
                IndexMatchUntyped(
                    SignalSimilarityInfoWithIntDistance(int(distance)),
                    self.local_id_to_entry[id][1],
                )
            )
        return matches

    def add(self, signal_str: str, entry: IndexT) -> None:
        self.add_all(((signal_str, entry),))

    def add_all(self, entries: t.Iterable[t.Tuple[str, IndexT]]) -> None:
        start = len(self.local_id_to_entry)
        self.local_id_to_entry.extend(entries)
        if start != len(self.local_id_to_entry):
            # This function signature is very silly
            self.index.add(
                (e[0] for e in self.local_id_to_entry[start:]),
                range(start, len(self.local_id_to_entry)),
            )


class PDQFlatIndex(PDQIndex):
    """
    Wrapper around the pdq faiss index lib
    that uses PDQFlatHashIndex instead of PDQMultiHashIndex
    It also uses a high match threshold to increase recall
    possibly as the cost of precision.
    """

    @classmethod
    def get_match_threshold(cls):
        return 52  # larger PDQ_MATCH_THRESHOLD for flatindexes

    @classmethod
    def _get_empty_index(cls) -> PDQHashIndex:
        return PDQFlatHashIndex()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from argparse import ArgumentParser, ArgumentTypeError
import argparse
import pathlib
import typing as t
from threatexchange.cli.helpers import FlexFilesInputAction
from threatexchange.exchanges.fetch_state import SignalOpinion, SignalOpinionCategory


from threatexchange.signal_type.signal_base import MatchesStr, SignalType, TextHasher

from threatexchange import common
from threatexchange.cli.cli_config import CLISettings
from threatexchange.content_type.content_base import ContentType
from threatexchange.exchanges.collab_config import CollaborationConfigBase
from threatexchange.cli import command_base


class LabelCommand(command_base.Command):
    """
    [WIP] Label signals and content for sharing.

    Warning! This command is still under construction, and is not yet stable.
    Please open an issue at https://github.com/facebook/ThreatExchange/issues
    if you need the development team to prioritize stabilitizing it.

    There are three main types of labeling:

    1. Seen
       Marking that you've observed a match, which can help others prioritize
       review, or track cross-platform spread.
    2. True Positive / False Positive
       After you've confirmed the results of a match, contributing what that
       result is can help others priotize signals with more precision.
    3. Upload
       If you have your own curated signals, sharing them with others can
       help them find matches, and give them opportunities to label your
       signals.


    Examples:
    ```
    $ threatexchange label "Sample Collab" text -l example,foo -- Some text I'm labeling
    ```
    """

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: ArgumentParser) -> None:
        ap.add_argument(
            "--labels",
            "-l",
            type=lambda s: set(s.strip().split(",")),
            metavar="CSV",
            default=set(),
            help="labels to apply to item",
        )

        signal_group = ap.add_mutually_exclusive_group()

        signal_group.add_argument(
            "--only-signals",
            "-S",
            nargs="+",
            type=common.argparse_choices_pre_type(
                [s.get_name() for s in settings.get_all_signal_types()],
                settings.get_signal_type,
            ),
            default=[],
            help="limit to this signal type",
        )

        signal_group.add_argument(
            "--as-hash",
            "-H",
            metavar="SIGNAL_TYPE",
            type=common.argparse_choices_pre_type(
                [s.get_name() for s in settings.get_all_signal_types()],
                settings.get_signal_type,
            ),
            help="interpret input as a hash of this type",
        )

        ap.add_argument(
            "collab",
            type=lambda n: _collab_type(n, settings),
            help="The name of the collaboration",
        )

        ap.add_argument(
            "content_type",
            type=common.argparse_choices_pre_type(
                [c.get_name() for c in settings.get_all_content_types()],
                settings.get_content_type,
            ),
            help="the type of what you are labeling",
        )

        ap.add_argument(
            "files",
            nargs=argparse.REMAINDER,
            action=FlexFilesInputAction,
            help="list of files or -- to interpret remainder as a string",
        )

    def __init__(
        self,
        content_type: t.Type[ContentType],
        files: t.List[pathlib.Path],
        as_hash: t.Optional[t.Type[SignalType]],
        collab: CollaborationConfigBase,
        only_signals: t.List[t.Type[SignalType]],
        labels: t.Set[str],
    ) -> None:
        self.collab = collab
        self.content_type = content_type
        self.files = files
        self.labels = labels
        self.only_signals = only_signals
        self.as_hash = as_hash

        if self.collab is None:
            raise ArgumentTypeError("No such collaboration!")

    def execute(self, settings: CLISettings) -> None:
        self.stderr("This command is not implemented yet, and most actions won't work")

        api = settings.apis.get_instance_for_collab(self.collab)
        # signal_types = self.only_signals or settings.get_signal_types_for_content(
        #     self.content_type
        # )

        if self.as_hash is not None:
            for f in self.files:
                signal_type = self.as_hash
                hash_val = signal_type.validate_signal_str(f.read_text())
                api.report_opinion(
                    signal_type,
                    hash_val,
                    SignalOpinion(
                        True, SignalOpinionCategory.POSITIVE_CLASS, self.labels
                    ),
                )
            return
        raise NotImplementedError


def _collab_type(name: str, settings: CLISettings) -> CollaborationConfigBase:
    ret = settings.get_collab(name)
    if ret is None:
        raise ArgumentTypeError(f"No such collab '{name}'!")
    return ret



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import collections
import csv
import pathlib
import re
import typing as t

from threatexchange.exchanges.clients.fb_threatexchange import threat_updates
from threatexchange.exchanges.clients.fb_threatexchange.descriptor import (
    SimpleDescriptorRollup,
)

_EXTENSION = ".te"

# TODO - merge SimpleDescriptorRollup here
class CliIndicatorSerialization(threat_updates.ThreatUpdateSerialization):
    """A short compact serialization optimized for the CLI"""

    def __init__(
        self,
        indicator_type: str,
        indicator: str,
        rollup: SimpleDescriptorRollup,
    ):
        self.indicator_type = indicator_type
        self.indicator = indicator
        self.rollup = rollup

    @property
    def key(self):
        return f"{self.indicator_type}.{self.indicator}"

    def as_csv_row(self) -> t.Tuple:
        """As a simple record type for the threatexchange CLI cache"""
        return (self.indicator,) + self.rollup.as_row()

    @classmethod
    def from_threat_updates_json(cls, app_id, te_json):
        return cls(
            te_json["type"],
            te_json["indicator"],
            SimpleDescriptorRollup.from_threat_updates_json(app_id, te_json),
        )

    @classmethod
    def te_threat_updates_fields(cls):
        return SimpleDescriptorRollup.te_threat_updates_fields()

    # ToDo this violates Liskov but is already used in Prod and will require a larger refactor
    @classmethod
    def store(
        cls, state_dir: pathlib.Path, contents: t.Iterable["CliIndicatorSerialization"]
    ) -> t.List[pathlib.Path]:
        # Stores in multiple files split by indicator type
        row_by_type = collections.defaultdict(list)
        for item in contents:
            row_by_type[item.indicator_type].append(item)
        ret = []
        for threat_type, items in row_by_type.items():
            path = state_dir / f"simple.{threat_type}{_EXTENSION}"
            ret.append(path)
            with path.open("w", encoding="utf-8", newline="") as f:
                writer = csv.writer(f)
                for item in items:
                    writer.writerow(item.as_csv_row())
        return ret

    @classmethod
    def load(cls, state_dir: pathlib.Path) -> t.Iterable["CliIndicatorSerialization"]:
        """Load this serialization from the state directory"""
        ret = []
        pattern = r"simple\.([^.]+)" + re.escape(_EXTENSION)
        for path in state_dir.glob(f"simple.*{_EXTENSION}"):
            match = re.match(pattern, path.name)
            if not match or not path.is_file():
                continue
            indicator_type = match.group(1)
            # Violate your warranty with class state! Not threadsafe!
            csv.field_size_limit(path.stat().st_size)  # dodge field size problems
            with path.open("r", encoding="utf-8", newline="") as f:
                for row in csv.reader(f):
                    ret.append(
                        cls(
                            indicator_type,
                            row[0],
                            SimpleDescriptorRollup.from_row(row[1:]),
                        )
                    )
        return ret


class HMASerialization(CliIndicatorSerialization):
    """
    A Serialization for HMA Similar to CliIndicatorSerialization but with
    Indicator ID.

    We also include the First Descriptor ID. The logic to determine which ID
    this is can be found in the SimpleDescriptorRollup
    """

    def __init__(
        self,
        indicator: str,
        indicator_type: str,
        indicator_id: str,
        rollup: SimpleDescriptorRollup,
    ):
        self.indicator_id = indicator_id
        self.indicator_type = indicator_type
        self.indicator = indicator
        self.rollup = rollup

    def as_csv_row(self) -> t.Tuple:
        """indicator details and descriptor rollup without descriptor ID"""
        return (self.indicator, self.indicator_id) + self.rollup.as_row()

    @classmethod
    def from_threat_updates_json(cls, app_id, te_json):
        return cls(
            te_json["indicator"],
            te_json["type"],
            te_json["id"],
            SimpleDescriptorRollup.from_threat_updates_json(app_id, te_json),
        )

    @classmethod
    def from_csv_row(
        cls, row: t.List[t.Any], indicator_type: str
    ) -> "HMASerialization":
        return cls(
            str(row[0]),
            indicator_type,
            str(row[1]),
            SimpleDescriptorRollup.from_row(row[2:]),
        )

    @classmethod
    def load(cls, state_dir: pathlib.Path) -> t.Iterable["HMASerialization"]:
        """Load this serialization from the state directory"""
        ret = []
        pattern = r"simple\.([^.]+)" + re.escape(_EXTENSION)
        for path in state_dir.glob(f"simple.*{_EXTENSION}"):
            match = re.match(pattern, path.name)
            if not match or not path.is_file():
                continue
            indicator_type = match.group(1)
            # Violate your warranty with class state! Not threadsafe!
            csv.field_size_limit(path.stat().st_size)  # dodge field size problems
            with path.open("r", newline="") as f:
                for row in csv.reader(f):
                    ret.append(cls.from_csv_row(row, indicator_type))
        return ret


if __name__ == "__main__":
    # Test Serialize Deserialize
    indicator = "indicator"
    indicator_id = "12345"
    first_descriptor_id = 6789
    added_on = "today"
    labels = {"tag1", "tag2"}

    ser = HMASerialization(
        indicator,
        "HASH_PDQ",
        indicator_id,
        SimpleDescriptorRollup(first_descriptor_id, added_on, labels),
    )
    serdeser = HMASerialization.from_csv_row(list(ser.as_csv_row()), "HASH_PDQ")

    if ser.as_csv_row() == serdeser.as_csv_row():
        print("Serialization worked correctly")
    else:
        print("Serialization failed")



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Config command to setup the CLI and settings.
"""

import argparse
import collections.abc
from dataclasses import is_dataclass, Field, fields, MISSING
from enum import Enum
import itertools
import json
import logging
import os
import typing as t

from threatexchange.exchanges.clients.fb_threatexchange.api import ThreatExchangeAPI
from threatexchange.extensions.manifest import ThreatExchangeExtensionManifest
from threatexchange import common, interface_validation
from threatexchange.cli.cli_config import CLISettings
from threatexchange.cli import command_base
from threatexchange.cli.exceptions import CommandError
from threatexchange.exchanges.impl.fb_threatexchange_api import (
    FBThreatExchangeCollabConfig,
    FBThreatExchangeCredentials,
    FBThreatExchangeSignalExchangeAPI,
)
from threatexchange.exchanges.signal_exchange_api import SignalExchangeAPI
from threatexchange.exchanges.impl.ncmec_api import NCMECSignalExchangeAPI
from threatexchange.exchanges.impl.static_sample import StaticSampleSignalExchangeAPI
from threatexchange.utils import dataclass_json


class ConfigCollabListCommand(command_base.Command):
    """List collaborations"""

    @classmethod
    def get_name(cls) -> str:
        return "list"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        # Ideas:
        #  * Filter by API
        pass

    def execute(self, settings: CLISettings) -> None:
        for collab in settings.get_all_collabs(default_to_sample=False):
            print(collab.api, collab.name)


class ConfigCollabPrintCommand(command_base.Command):

    """
    Print a collab config to screen.

    This can be used to share configs via creation from JSON:

    ```
    $ threatexchange config collab print $name > file
    $ cat file
    {
        "filename": "/local/file.txt",
        "name": "my_collab",
        "api": "local_file",
        "enabled": true,
        "signal_type": null
    }
    $ threatexchange config collab delete my_collab
    $ threatexchange config collab edit local_file --create --json "$(cat file)"
    ```
    """

    @classmethod
    def get_name(cls) -> str:
        return "print"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        ap.add_argument("collab_name", help="the name of the collab")

    def __init__(
        self,
        collab_name: str,
    ) -> None:
        self.collab_name = collab_name

    def execute(self, settings: CLISettings) -> None:
        collab = settings.get_collab(self.collab_name)
        if collab is None:
            raise CommandError.user(f"No such collab {self.collab_name}")
        print(dataclass_json.dataclass_dumps(collab))


class _UpdateCollabCommand(command_base.Command):
    """
    Create or edit collaborations for this API

    Programatically generated by inspecting the config class, so not everything will be
    documented.
    """

    _API_CLS: t.ClassVar[t.Type[SignalExchangeAPI]]

    _IGNORE_FIELDS = {
        "name",
        "api",
        "enabled",
        # "only_signal_types",
        # "not_signal_types",
        # "only_owners",
        # "not_owners",
        # "only_tags",
        # "not_tags",
    }

    @classmethod
    def get_name(cls) -> str:
        return cls._API_CLS.get_name()

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        cfg_cls = cls._API_CLS.get_config_cls()
        assert is_dataclass(cfg_cls)

        ap.add_argument("collab_name", help="the name of the collab")
        ap.set_defaults(api_name=cls._API_CLS.get_name())
        ap.add_argument(
            "--create",
            "-C",
            action="store_true",
            help="indicate you intend to create a config",
        )

        cfg_common_ap = ap.add_argument_group(description="common to all collabs")
        on_off = cfg_common_ap.add_mutually_exclusive_group()

        # This goofy syntax allows --enable, --enable=1, and enable=0 to disable
        on_off.add_argument(
            "--enable",
            nargs="?",
            type=int,
            const=1,
            choices=[0, 1],
            help="enable the config (default on create)",
        )
        on_off.add_argument(
            "--disable",
            dest="enable",
            action="store_const",
            const=0,
            help="disable the config",
        )

        type_specific_fields = [f for f in fields(cfg_cls) if cls._is_argument_field(f)]
        if type_specific_fields:
            config_ap = ap.add_argument_group(
                description=f"specific to {cls._API_CLS.get_name()}"
            )
            for field in type_specific_fields:
                cls._add_argument(config_ap, field)

        ap.add_argument(
            "--json",
            "-J",
            dest="is_json",
            action="store_true",
            help="instead, interpret the argument as JSON and use that to edit the config",
        )

    @classmethod
    def _is_argument_field(cls, field: Field) -> bool:
        if not field.init:
            return False
        if field.name in cls._IGNORE_FIELDS:
            return False
        return True

    @classmethod
    def _add_argument(cls, ap: argparse._ArgumentGroup, field: Field) -> None:
        assert cls._is_argument_field(field)
        assert not isinstance(
            field.type, t.ForwardRef
        ), "rework class to not have forward ref"

        origin = t.get_origin(field.type)
        argparse_type: t.Callable[[str], t.Any] = field.type
        metavar: str
        if isinstance(field.type, type) and issubclass(field.type, Enum):
            argparse_type = common.argparse_choices_pre_type(
                [m.name for m in field.type],
                lambda s: field.type[s],
            )
            metavar = f"[{','.join(m.name for m in field.type)}]"
        elif origin is not None:
            arg_type = t.get_args(field.type)[0]
            if isinstance(origin, type) and issubclass(
                origin, collections.abc.Collection
            ):
                argparse_type = lambda s: origin(arg_type(p.strip()) for p in s.split(","))  # type: ignore  # mypy not smart enough for origin == type
                metavar = f"{arg_type.__name__}[,{arg_type.__name__}[,...]]"
            elif origin == t.Union:  # Should this be is?
                argparse_type = arg_type
                metavar = arg_type.__name__
            else:
                raise AssertionError(
                    f"Unhandled complex type for {field.name}: {field.type}"
                )
        else:
            metavar = field.type.__name__

        help = "[missing] Add a help annotation on the config class!"
        if field.metadata:
            metavar = field.metadata.get("metavar", metavar)
            help = field.metadata.get("help", help)

        ap.add_argument(
            f"--{field.name.replace('_', '-')}",
            type=argparse_type,
            metavar=metavar,
            required=field.default is MISSING and field.default_factory is MISSING,
            help=help,
        )

    def __init__(
        self,
        full_argparse_namespace,
        create: bool,
        collab_name: str,
        enable: t.Optional[int],
        is_json: bool,
    ) -> None:
        self.namespace = full_argparse_namespace
        self.create = create
        self.edit_kwargs = {}
        self.collab_name = collab_name
        if is_json:
            self.edit_kwargs = json.loads(collab_name)
            self.collab_name = self.edit_kwargs["name"]

        # Technically you could combine the flags and JSON, but you'd be weird
        if create:
            self.edit_kwargs["name"] = self.collab_name
            self.edit_kwargs["enabled"] = True
            self.edit_kwargs["api"] = self._API_CLS.get_name()

        if enable is not None:
            self.edit_kwargs["enabled"] = bool(enable)

        for field in fields(self._API_CLS.get_config_cls()):
            if not field.init:
                if field.name == "api":
                    self.edit_kwargs.pop("api", None)
                continue
            if field.name in self._IGNORE_FIELDS:
                continue
            val = getattr(full_argparse_namespace, field.name)
            if val is not None:
                self.edit_kwargs[field.name] = val

    def execute(self, settings: CLISettings) -> None:
        existing = settings.get_collab(self.collab_name)

        if existing:
            if self.create:
                raise CommandError(
                    f'there\'s an existing collaboration named "{self.collab_name}"', 2
                )
            if existing.api != self._API_CLS.get_name():
                raise CommandError(
                    f"the existing collab is for the {existing.api} api, delete that one first",
                    2,
                )
            assert (
                existing.__class__ == self._API_CLS.get_config_cls()
            ), "api name the same, but class different?"
            for name, val in self.edit_kwargs.items():
                setattr(existing, name, val)
            settings._state.update_collab(existing)
        elif self.create:
            logging.debug("Creating config with args: %s", self.edit_kwargs)
            to_create = self._API_CLS.get_config_cls()(**self.edit_kwargs)
            settings._state.update_collab(to_create)
        else:
            raise CommandError("no such config! Did you mean to use --create?", 2)


class ConfigCollabForAPICommand(command_base.CommandWithSubcommands):
    """Create and edit collaborations for APIs"""

    @classmethod
    def get_name(cls) -> str:
        return "edit"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        cls._SUBCOMMANDS = [
            cls._create_command_for_api(api)
            for api in settings.apis
            if api.__class__ is not StaticSampleSignalExchangeAPI
        ]

    @classmethod
    def _create_command_for_api(
        cls, api: t.Type[SignalExchangeAPI]
    ) -> t.Type[command_base.Command]:
        """Don't try this at home!"""

        class _GeneratedUpdateCommand(_UpdateCollabCommand):
            _API_CLS = api

        _GeneratedUpdateCommand.__name__ = (
            f"{_GeneratedUpdateCommand.__name__}_{api.get_name()}"
        )

        return _GeneratedUpdateCommand


class ConfigCollabDeleteCommand(command_base.Command):
    """Delete collaborations"""

    @classmethod
    def get_name(cls) -> str:
        return "delete"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        ap.add_argument("collab_name", help="the collab to delete")

    def __init__(self, collab_name: str) -> None:
        self.collab_name = collab_name

    def execute(self, settings: CLISettings) -> None:
        collab = settings.get_collab(self.collab_name)
        if collab is None:
            raise CommandError("No such collab", 2)
        settings._state.delete_collab(collab)  # TODO clean private member access


class ConfigCollabCommand(command_base.CommandWithSubcommands):
    """Configure collaborations"""

    _SUBCOMMANDS = [
        ConfigCollabListCommand,
        ConfigCollabPrintCommand,
        ConfigCollabForAPICommand,
        ConfigCollabDeleteCommand,
    ]

    @classmethod
    def get_name(cls) -> str:
        return "collab"

    def execute(self, settings: CLISettings) -> None:
        ConfigCollabListCommand().execute(settings)


class ConfigExtensionsCommand(command_base.Command):
    """Configure extensions"""

    @classmethod
    def get_name(cls) -> str:
        return "extensions"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        ap.add_argument(
            "action",
            choices=["list", "add", "remove"],
            default="list",
            help="what to do",
        )
        ap.add_argument(
            "module",
            nargs="?",
            help="the module path to the extension. foo.bar.baz",
        )
        ap.set_defaults(is_config=True)

    def __init__(self, action: str, module: t.Optional[str]) -> None:
        self.action = {
            "list": self.execute_list,
            "add": self.execute_add,
            "remove": self.execute_remove,
        }[action]
        self.module = module

    def execute(self, settings: CLISettings) -> None:
        self.action(settings)

    def execute_list(self, settings: CLISettings) -> None:
        if self.module:
            manifest = self.get_manifest(self.module)
            self.print_extension(manifest)
            return
        for module_name in sorted(settings.get_persistent_config().extensions):
            print(module_name)
            manifest = self.get_manifest(module_name)
            self.print_extension(manifest, indent=2)

    def get_manifest(self, module_name: str) -> ThreatExchangeExtensionManifest:
        try:
            return ThreatExchangeExtensionManifest.load_from_module_name(module_name)
        except ValueError as ve:
            raise CommandError(str(ve), 2)

    def execute_add(self, settings: CLISettings) -> None:
        if not self.module:
            raise CommandError("module is required", 2)
        if self.module in settings.get_persistent_config().extensions:
            self.execute_list(settings)
            return

        manifest = self.get_manifest(self.module)

        # Validate our new setups by pretending to create a new mapping with the new classes
        content_and_settings = interface_validation.SignalTypeMapping(
            list(
                itertools.chain(
                    settings.get_all_content_types(), manifest.content_types
                )
            ),
            list(
                itertools.chain(settings.get_all_signal_types(), manifest.signal_types)
            ),
        )

        apis = list(manifest.apis)
        apis.extend(settings.apis.get_all())
        interface_validation.SignalExchangeAPIMapping(apis)

        self.execute_list(settings)

        config = settings.get_persistent_config()
        config.extensions.add(self.module)
        settings.set_persistent_config(config)

    def execute_remove(self, settings: CLISettings) -> None:
        if not self.module:
            raise CommandError("Which module you are remove is required", 2)
        config = settings.get_persistent_config()
        if self.module not in config.extensions:
            raise CommandError(f"You haven't added {self.module}", 2)
        config.extensions.remove(self.module)
        settings.set_persistent_config(config)

    def print_extension(
        self, manifest: ThreatExchangeExtensionManifest, indent=0
    ) -> None:
        space = " " * indent
        level2 = f"\n{space}  "
        if manifest.signal_types:
            print(f"{space}Signal:{level2}", end="")
            print(
                level2.join(
                    f"{s.get_name()} - {s.__name__}" for s in manifest.signal_types
                )
            )
        if manifest.content_types:
            print(f"{space}Content:{level2}", end="")
            print(
                level2.join(
                    f"{c.get_name()} - {c.__name__}" for c in manifest.content_types
                )
            )
        if manifest.apis:
            print(f"{space}Content:{level2}", end="")
            print(level2.join(f"{a.get_name()} - {a.__name__}" for a in manifest.apis))


class ConfigSignalCommand(command_base.Command):
    """Configure and view available SignalTypes"""

    @classmethod
    def get_name(cls) -> str:
        return "signal"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        ap.add_argument(
            "action",
            choices=["list"],
            nargs="?",
            default="list",
            help="what to do",
        )

    def __init__(self, action: str) -> None:
        self.action = {
            "list": self.execute_list,
        }[action]

    def execute(self, settings: CLISettings) -> None:
        self.action(settings)

    def execute_list(self, settings: CLISettings) -> None:
        signals = settings.get_all_signal_types()
        for name, class_name in sorted(
            (st.get_name(), _fully_qualified_name(st)) for st in signals
        ):
            print(name, class_name)


class ConfigContentCommand(command_base.Command):
    """Configure and view available ContentTypes"""

    @classmethod
    def get_name(cls) -> str:
        return "content"

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        ap.add_argument(
            "action",
            choices=["list"],
            nargs="?",
            default="list",
            help="what to do",
        )

    def __init__(self, action: str) -> None:
        self.action = {
            "list": self.execute_list,
        }[action]

    def execute(self, settings: CLISettings) -> None:
        self.action(settings)

    def execute_list(self, settings: CLISettings) -> None:
        content_types = settings.get_all_content_types()
        for name, class_name in sorted(
            (c.get_name(), _fully_qualified_name(c)) for c in content_types
        ):
            print(name, class_name)


class ConfigThreatExchangeAPICommand(command_base.Command):
    """Configure Facebook ThreatExchange integration"""

    @classmethod
    def get_name(cls) -> str:
        return FBThreatExchangeSignalExchangeAPI.get_name()

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        import_cmds = ap.add_mutually_exclusive_group()
        import_cmds.add_argument(
            "--list-available-collabs",
            "-L",
            action="store_true",
            help="query the API to list available collabs",
        )
        import_cmds.add_argument(
            "--import-collab",
            "-I",
            type=int,
            help="import a collaboration by privacy group ID",
        )

        # Not actually a type of import cmd, but to add exclusivity logic
        import_cmds.add_argument(
            "--api-token",
            help="set the default api token (https://developers.facebook.com/tools/accesstoken/)",
        )

    def __init__(
        self,
        api_token: t.Optional[str],
        list_available_collabs: bool,
        import_collab: t.Optional[int],
    ) -> None:
        self.api_token = api_token
        self.action = self.execute_config
        if list_available_collabs:
            self.action = self.execute_list_collabs
        elif import_collab is not None:
            tmp_for_typing = import_collab
            self.action = lambda s: self.execute_import(s, tmp_for_typing)

    def execute(self, settings: CLISettings) -> None:
        self.action(settings)

    def get_te_api(self) -> ThreatExchangeAPI:
        creds = FBThreatExchangeCredentials.get(FBThreatExchangeSignalExchangeAPI)
        return ThreatExchangeAPI(creds.api_token)

    def execute_list_collabs(self, settings: CLISettings) -> None:
        api = self.get_te_api()

        unique_privacy_groups = {
            pg.id: pg for pg in api.get_threat_privacy_groups_member()
        }
        unique_privacy_groups.update(
            (pg.id, pg) for pg in api.get_threat_privacy_groups_owner()
        )

        max_width = os.get_terminal_size().columns

        for pg in sorted(unique_privacy_groups.values(), key=lambda pg: pg.name):
            if not pg.threat_updates_enabled:
                continue
            line = f"{pg.id} {pg.name} - {pg.description}".replace("\n", " ")
            if len(line) > max_width:
                line = f"{line[:max_width-3]}..."
            print(line)

    def execute_import(self, settings: CLISettings, privacy_group_id: int) -> None:
        api = self.get_te_api()
        pg = api.get_privacy_group(privacy_group_id)
        if settings.get_collab(pg.name) is not None:
            raise CommandError(
                f"A collaboration already exists with the name {pg.name}", 2
            )
        settings._state.update_collab(
            FBThreatExchangeCollabConfig(name=pg.name, privacy_group=privacy_group_id)
        )

    def execute_config(self, settings: CLISettings) -> None:
        if self.api_token is not None:
            config = settings.get_persistent_config()
            config.fb_threatexchange_api_token = self.api_token
            settings.set_persistent_config(config)


class ConfigNCMECAPICommand(command_base.Command):
    """Configure NCMEC hash api integration"""

    @classmethod
    def get_name(cls) -> str:
        return NCMECSignalExchangeAPI.get_name()

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:
        ap.add_argument(
            "--credentials",
            metavar="STR",
            nargs=2,
            help="set the username and password to access the NCMEC API",
        )

    def __init__(
        self,
        credentials: t.List[str],
    ) -> None:
        self.credentials = (credentials[0], credentials[1]) if credentials else None

    def execute(self, settings: CLISettings) -> None:
        if self.credentials is not None:
            config = settings.get_persistent_config()
            config.ncmec_credentials = self.credentials
            settings.set_persistent_config(config)


class ConfigAPICommand(command_base.CommandWithSubcommands):
    """Configure and view available SignalExchangeAPIs"""

    _SUBCOMMANDS = [ConfigThreatExchangeAPICommand, ConfigNCMECAPICommand]

    @classmethod
    def get_name(cls) -> str:
        return "api"

    def execute(self, settings: CLISettings) -> None:
        apis = settings.apis.get_all()
        for name in sorted(a.get_name() for a in apis):
            print(name)


class ConfigCommand(command_base.CommandWithSubcommands):
    """Configure the CLI"""

    _SUBCOMMANDS = [
        ConfigCollabCommand,
        ConfigSignalCommand,
        ConfigContentCommand,
        ConfigAPICommand,
        ConfigExtensionsCommand,
    ]


def _fully_qualified_name(klass: t.Type):
    return f"{klass.__module__}.{klass.__qualname__}"



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from threatexchange.signal_type.tests.signal_type_test_helper import MatchesStrAutoTest

from threatexchange.signal_type.raw_text import RawTextSignal


class TestRawTextSignal(MatchesStrAutoTest):

    TYPE = RawTextSignal

    def get_validate_hash_cases(self):
        return [
            ("a", "a"),
            ("a ", "a"),
        ]

    def get_compare_hash_cases(self):
        return []

    def get_matches_str_cases(self):
        return [
            ("", ""),
            ("a", "a"),
            ("a", "b", False, 1),
            ("aaaaaaaaaa", "a", False, 9),
            ("a", "aaaaaaaaa", False, 8),
            # Normalization removes spaces
            ("a a a a a a a a a", "aaaaaaaaa", True, 0),
            # Default threshold is 95%
            ("a" * 19, "a" * 18 + "b", False, 1),
            ("a" * 20, "a" * 19 + "b", True, 1),
        ]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import pytest
from pathlib import Path

try:
    import vpdq

    _DISABLED = False
except (ImportError, ModuleNotFoundError) as e:
    _DISABLED = True
else:
    from threatexchange.extensions.vpdq.vpdq import VPDQSignal
    from threatexchange.extensions.vpdq.vpdq_util import (
        vpdq_to_json,
        json_to_vpdq,
        VpdqCompactFeature,
        VPDQ_QUALITY_THRESHOLD,
        VPDQ_DISTANCE_THRESHOLD,
    )
    from threatexchange.extensions.vpdq.vpdq_brute_matcher import (
        match_VPDQ_hash_brute,
    )

VIDEO = "tmk/sample-videos/chair-20-sd-bar.mp4"
ROOTDIR = Path(__file__).parents[5]


@pytest.mark.skipif(_DISABLED, reason="vpdq not installed")
def test_vpdq_from_string_path():
    computed_hash = VPDQSignal.hash_from_file(ROOTDIR / VIDEO)
    expected_hash = [
        VpdqCompactFeature(h, 100, float(i))
        for i, h in enumerate(
            (
                "e42926b9cb58d261383b699f0ff2e1076fe6183dff4986d82ca36da48034fc54",
                "4ce949b11f1035e2f2eb0a5e76f68716870e70f97bd91d1049a3dbe1d164b454",
                "643c8f9e5f0eb6866886355a4a4bdbe38343a4e924b969b41d321ba33a85f55e",
                "02b456d6bd4b6d695aa51772cdb8fa1df0ced12ea64b82f106e53d1b7e0b8050",
                "e64c35684e8d8926cc96324b66b9c999d92c37566669999b9bb166e464cbdd11",
                "46d86a5cc1ac1787f2b5cb895833c6a59e07b37ca60524ccaf1b69af9b2d9053",
                "c5f2ada74d0649858d6f1d26adace8fcd4b98d1938d131fca75a9611b41332d2",
                "093455b85518971cb48dad0f2b464a7552b3049345a905e904db2fde6ff66ff9",
                "e25339b25ecca73c5b1c84650a4b17a6afd21eac3d12fa14f00ff81bf03ce178",
                "8834b4ce2ca75b4d964e2526096f02cc27a63fa7be42fc16f83f486b487e4b78",
                "b26d60de1b62b499e0bc83680f8c3b25f0fae3b6c72b6e397c3414710e634d4e",
                "b506c921960c6d24e70705071c3b64e6e1a7ced61cf639ea199e1e345b726f6e",
                "93457d84eb67a4c33a63f4ca9307361b7161e709cc1578c129c93cbc94b0df5e",
                "419d81cc6926e1c6cfe67ae6e1e6c25b9af03349e7057641acc98c1c96196f5e",
                "64368799091de307e9855b6ee1e6c3e61ff633618701c64378c91c9d9499d75a",
                "256393c7dbcc7cb93665831385dae732586338abf439c89683c6a3e1d129b552",
                "cad9719334499e3dcb3361233744e9cc519b2a89d7e8a21734f679696921d556",
                "e0d8834b1f0bebf1e362e4ad0bac1b5efc9a81cb6ee97c34812793a19285e554",
                "9d4a35397761e5511559f178e5660f06dd4ec1b97f99b952122326e9e269d114",
                "792efd6e85e31b812a5162f95ea63a5622d60bb99d70a2564aab5de99549a154",
                "efc4c1861a3326f148d9de093bb9c1b65d46acab80dcdf42895e223d7f61c554",
            )
        )
    ]
    assert vpdq_to_json(json_to_vpdq(computed_hash)) == computed_hash
    res = match_VPDQ_hash_brute(
        json_to_vpdq(computed_hash),
        expected_hash,
        VPDQ_QUALITY_THRESHOLD,
        VPDQ_DISTANCE_THRESHOLD,
    )
    assert res.query_match_percent == 100
    assert res.compared_match_percent == 100



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from threatexchange.extensions.manifest import ThreatExchangeExtensionManifest
from threatexchange.extensions.vpdq.vpdq import VPDQSignal


TX_MANIFEST = ThreatExchangeExtensionManifest(
    signal_types=(VPDQSignal,),
)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
SignalExchangeAPI impl for the NCMEC hash exchange API

@see NCMECSignalExchangeAPI
"""


import logging
import time
import typing as t
from dataclasses import dataclass, field

from threatexchange.exchanges.clients.ncmec import hash_api as api

from threatexchange.exchanges import auth, fetch_state as state
from threatexchange.exchanges import signal_exchange_api
from threatexchange.exchanges.collab_config import (
    CollaborationConfigWithDefaults,
)
from threatexchange.signal_type.signal_base import SignalType
from threatexchange.signal_type.md5 import VideoMD5Signal
from threatexchange.signal_type.pdq.signal import PdqSignal


_API_NAME: str = "ncmec"


@dataclass
class NCMECCheckpoint(
    state.FetchCheckpointBase,
):
    """
    NCMEC primarily revolves around polling the timestamp.

    NCMEC IDs seem to stay around forever, so no need for is_stale()
    """

    # The biggest value of "to", and the next "from"
    get_entries_max_ts: int

    def get_progress_timestamp(self) -> t.Optional[int]:
        return self.get_entries_max_ts

    @classmethod
    def from_ncmec_fetch(cls, response: api.GetEntriesResponse) -> "NCMECCheckpoint":
        return cls(response.max_timestamp)

    def __setstate__(self, d: t.Dict[str, t.Any]) -> None:
        """Implemented for pickle version compatibility."""
        # 0.99.0 => 1.0.0:
        ### field 'max_timestamp' renamed to 'get_entries_max_ts'
        if "max_timestamp" in d:
            d["get_entries_max_ts"] = d.pop("max_timestamp")
        self.__dict__ = d


@dataclass
class _NCMECCollabConfigRequiredFields:
    environment: api.NCMECEnvironment = field(
        metadata={"help": "which database to connect to"}
    )


@dataclass
class NCMECCollabConfig(
    CollaborationConfigWithDefaults,
    _NCMECCollabConfigRequiredFields,
):
    api: str = field(init=False, default=_API_NAME)
    only_esp_ids: t.Set[int] = field(
        default_factory=set,
        metadata={
            "help": "Only take entries from these eletronic service provider (ESP) ids"
        },
    )


@dataclass
class NCMECOpinion(state.SignalOpinion):
    esp_id: int

    def __setstate__(self, d: t.Dict[str, t.Any]) -> None:
        """Implemented for pickle version compatibility."""
        # 0.99.0 => 1.0.0:
        ### field 'owner_id' renamed to 'esp_id'
        if "owner" in d:
            d["esp_id"] = d["owner"]
        super().__setstate__(d)


@dataclass
class NCMECSignalMetadata(state.FetchedSignalMetadata):
    """
    NCMEC metadata includes who uploaded it, as well as what they tagged.

    The NCMEC API has no concept of false positives - every entry is reported.
    """

    member_entries: t.Dict[int, t.Set[str]]

    def get_as_opinions(self) -> t.Sequence[NCMECOpinion]:
        return [
            NCMECOpinion(
                False,  # TODO - get my own esp_id
                state.SignalOpinionCategory.POSITIVE_CLASS,
                tags,
                member_id,
            )
            for member_id, tags in self.member_entries.items()
        ]


def _get_conversion(
    signal_types: t.Sequence[t.Type[SignalType]],
) -> t.Mapping[t.Tuple[api.NCMECEntryType, str], t.Type[SignalType]]:
    ret: t.Dict[t.Tuple[api.NCMECEntryType, str], t.Type[SignalType]] = {}
    if VideoMD5Signal in signal_types:
        ret[api.NCMECEntryType.video, "md5"] = VideoMD5Signal
    if PdqSignal in signal_types:
        ret[api.NCMECEntryType.image, "pdq"] = PdqSignal
    for st in signal_types:
        if st.get_name() == "photodna":
            ret[api.NCMECEntryType.image, "pdna"] = st
            break
    return ret


@dataclass
class NCMECCredentials(auth.CredentialHelper):
    ENV_VARIABLE: t.ClassVar[str] = "TX_NCMEC_CREDENTIALS"
    FILE_NAME: t.ClassVar[str] = "~/.tx_ncmec_credentials"

    user: str
    password: str

    @classmethod
    def _from_str(cls, s: str) -> "NCMECCredentials":
        user, _, passw = s.strip().partition(":")
        return cls(user, passw)

    def _are_valid(self) -> bool:
        return bool(self.user and self.password)


class NCMECSignalExchangeAPI(
    auth.SignalExchangeWithAuth[NCMECCollabConfig, NCMECCredentials],
    signal_exchange_api.SignalExchangeAPI[
        NCMECCollabConfig,
        NCMECCheckpoint,
        NCMECSignalMetadata,
        str,
        api.NCMECEntryUpdate,
    ],
):
    """
    Conversion for the NCMEC hash API

    Key implementation details:
        1. API is a stream of content: opinion, hashes,
           which need to be remapped to hash => opinion
        2. Owners have ids
        3. As of 5/2022 there are no false positive or seen statuses
    """

    MAX_FETCH_SIZE: t.ClassVar[int] = 400000
    FETCH_SHRINK_FACTOR: t.ClassVar[int] = 4

    def __init__(
        self,
        collab: NCMECCollabConfig,
        username: str,
        password: str,
    ) -> None:
        super().__init__()
        self.collab = collab
        self._username = username
        self._password = password

    @classmethod
    def for_collab(
        cls,
        collab: NCMECCollabConfig,
        credentials: t.Optional["NCMECCredentials"] = None,
    ) -> "NCMECSignalExchangeAPI":
        credentials = credentials or NCMECCredentials.get(cls)
        return cls(collab, credentials.user, credentials.password)

    @classmethod
    def get_name(cls) -> str:
        return _API_NAME

    @staticmethod
    def get_config_cls() -> t.Type[NCMECCollabConfig]:
        return NCMECCollabConfig

    @staticmethod
    def get_checkpoint_cls() -> t.Type[NCMECCheckpoint]:
        return NCMECCheckpoint

    @staticmethod
    def get_record_cls() -> t.Type[NCMECSignalMetadata]:
        return NCMECSignalMetadata

    @staticmethod
    def get_credential_cls() -> t.Type[NCMECCredentials]:
        return NCMECCredentials

    def get_client(self, environment: api.NCMECEnvironment) -> api.NCMECHashAPI:
        if not api.is_valid_user_pass(self._username, self._password):
            raise Exception("NCMEC username and password not configured or invalid.")
        return api.NCMECHashAPI(self._username, self._password, environment)

    def fetch_iter(
        self,
        _supported_signal_types: t.Sequence[t.Type[SignalType]],
        checkpoint: t.Optional[NCMECCheckpoint],
    ) -> t.Iterator[state.FetchDelta[str, api.NCMECEntryUpdate, NCMECCheckpoint]]:
        """
        Use flow control to efficiently fetch from the NCMEC API.

        The NCMEC API does not provide entries in a strictly ascending time order.
        As a result, the only checkpoint we can safely return is if we have
        exhausted an entire time range of fetching.

        However, the data within the NCMEC hash dbs are randomly distributed,
        with the following observed behavior:
            1. Fetching an empty range (0 entries) is fast
            2. We can estimate the number of entries in a range based on the
               parameters of the "next" field
            3. Fetching from a cursor of data is significantly faster than
               generating a new cursor.

        Therefore, we can aim for the following strategy:
            1. Prefer fetching empty ranges repeatedly rather than potentially
               generating an overfetch. (Shrink quickly, grow slowly)
            2. Be very generous with overfetch limits, since if we've generated
               the cursor
        """
        start_time = 0
        if checkpoint is not None:
            start_time = checkpoint.get_entries_max_ts
        # Avoid being exactly at end time for updates showing up multiple
        # times in the fetch, since entries are not ordered by time
        end_time = int(time.time()) - 5

        client = self.get_client(self.collab.environment)
        # We could probably mutate start time, but new variable for clarity
        current_start = start_time
        # The range we are fetching
        duration = end_time - current_start
        # A counter for when we want to increase our duration
        # We want to be conservative
        low_fetch_counter = 0

        def log(event: str) -> None:
            """Helper to log fetch events"""
            if duration < 1800:
                duration_str = f"{duration} seconds"
            elif duration < 43200:
                duration_str = f"{duration / 3600:.2f} hours"
            else:
                duration_str = f"{duration / (3600 * 24):.2f} days"
            logging.info(
                "NCMEC API %s @%s (%s)",
                event,
                api._date_format(current_start),
                duration_str,
            )

        while current_start < end_time:
            duration = max(1, duration)  # Infinite loop defense
            # Don't fetch past our designated end
            current_end = min(end_time, current_start + duration)
            updates: t.List[api.NCMECEntryUpdate] = []
            for i, entry in enumerate(
                client.get_entries_iter(
                    start_timestamp=current_start, end_timestamp=current_end
                )
            ):
                if i == 0:  # First batch, check for overfetch
                    if (
                        entry.estimated_entries_in_range > self.MAX_FETCH_SIZE
                        and duration > 1
                    ):
                        log(
                            f"est {entry.estimated_entries_in_range} is over max fetch, duration reduced"
                        )
                        # We want to at last shrink by our shrink factor
                        duration = min(
                            duration // self.FETCH_SHRINK_FACTOR,
                            # Especially in early fetches, we are overfetching
                            # by a huge amount, so shrink in proportion to overfetch
                            duration
                            * self.MAX_FETCH_SIZE
                            // entry.estimated_entries_in_range,
                        )
                        low_fetch_counter = 0  # Don't grow right after a shrink
                        break  # Retry get_entries_iter with new parameters
                    else:
                        # Our entry estimatation (based on the cursor parameters)
                        # occasionally seem to over-estimate
                        log(f"est {entry.estimated_entries_in_range} entries")
                elif i % 100 == 0:
                    # If we get down to one second, we can potentially be
                    # fetching an arbitrary large amount of data in one go,
                    # so log something occasionally
                    log(f"large fetch ({i}), up to {len(updates)}")
                updates.extend(entry.updates)
            else:  # AKA a successful fetch
                # If we're hovering near the single-fetch limit for a period
                # of time, we can likely safely expand our range.
                if len(updates) < api.NCMECHashAPI.ENTRIES_PER_FETCH * 2:
                    low_fetch_counter += 1
                    if low_fetch_counter >= self.FETCH_SHRINK_FACTOR:
                        log("multiple low fetches, increasing duration")
                        # Grow slower than we shrink
                        duration *= self.FETCH_SHRINK_FACTOR // 2
                        low_fetch_counter = 0
                # If we are not quite at our limit, but getting close to it,
                # pre-emptively shrink to try and stay under the limit
                elif len(updates) > self.MAX_FETCH_SIZE / self.FETCH_SHRINK_FACTOR:
                    log("close to overfetch limit, reducing duration")
                    duration //= self.FETCH_SHRINK_FACTOR
                    low_fetch_counter = 0
                else:  # Not too small, not too large, just right
                    low_fetch_counter = 0
                yield state.FetchDelta(
                    {f"{entry.member_id}-{entry.id}": entry for entry in updates},
                    NCMECCheckpoint(current_end),
                )
                current_start = current_end

    @classmethod
    def fetch_value_merge(
        cls,
        old: t.Optional[api.NCMECEntryUpdate],
        new: t.Optional[api.NCMECEntryUpdate],
    ) -> t.Optional[api.NCMECEntryUpdate]:
        assert new is not None, "fetch shouldn't do this"
        if new.deleted:
            return None
        return new

    @classmethod
    def naive_convert_to_signal_type(
        cls,
        signal_types: t.Sequence[t.Type[SignalType]],
        collab: NCMECCollabConfig,
        fetched: t.Mapping[str, api.NCMECEntryUpdate],
    ) -> t.Dict[t.Type[SignalType], t.Dict[str, NCMECSignalMetadata]]:
        mapping: t.Mapping[
            t.Tuple[api.NCMECEntryType, str], t.Type[SignalType]
        ] = _get_conversion(signal_types)
        ret: t.Dict[t.Type[SignalType], t.Dict[str, NCMECSignalMetadata]] = {}
        for entry in fetched.values():
            if entry.deleted:
                continue  # We expect len(fingerprints) == 0 here, but to be safe
            if collab.only_esp_ids and entry.member_id not in collab.only_esp_ids:
                continue
            for fingerprint_type, fingerprint_value in entry.fingerprints.items():
                st = mapping.get((entry.entry_type, fingerprint_type))
                if st is not None:
                    try:
                        signal_value = st.validate_signal_str(fingerprint_value)
                    except Exception:
                        logging.warning(
                            "Invalid fingerprint (%s): %s",
                            st.get_name(),
                            fingerprint_value,
                        )
                        continue
                    metadata = ret.setdefault(st, {}).setdefault(
                        signal_value, NCMECSignalMetadata({})
                    )
                    tags = metadata.member_entries.setdefault(entry.member_id, set())
                    if entry.classification:
                        tags.add(entry.classification)
        return ret



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import pytest
import sys
import typing as t

from _pytest.config import Config

from threatexchange.cli import main
from threatexchange.cli.command_base import Command, CommandWithSubcommands
from threatexchange.cli.tests.e2e_test_helper import ThreatExchangeCLIE2eHelper, te_cli

_PRINT_TO_STDERR: bool = False


def test_all_helps(te_cli: ThreatExchangeCLIE2eHelper, pytestconfig: Config):
    """
    Just executes all the commands to make sure they don't throw on help.

    View the pretty output with py.test -sv
    """
    global _PRINT_TO_STDERR
    _PRINT_TO_STDERR = pytestconfig.getoption("verbose") > 0
    te_cli.cli_call()  # root help
    _run_help(te_cli)  # root help
    for command in main.get_subcommands():
        _recurse(te_cli, command)


def _run_help(te_cli: ThreatExchangeCLIE2eHelper, *args: str) -> None:
    out = te_cli.cli_call(*args, "--help")
    if _PRINT_TO_STDERR:
        print(out, file=sys.stderr)


def _recurse(
    te_cli: ThreatExchangeCLIE2eHelper, command: t.Type[Command], *parents: str
) -> None:
    name = command.get_name()
    _run_help(te_cli, *parents, name)
    if issubclass(command, CommandWithSubcommands):
        for subcommand in command._SUBCOMMANDS:
            _recurse(te_cli, subcommand, *parents, name)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""SignalExchangeAPI implementation for StopNCII.org"""


from functools import lru_cache
import time
import typing as t
from dataclasses import dataclass

from threatexchange.exchanges.clients.stopncii import api

from threatexchange.exchanges import fetch_state as state
from threatexchange.exchanges import signal_exchange_api
from threatexchange.exchanges import auth
from threatexchange.exchanges.collab_config import (
    CollaborationConfigBase,
)
from threatexchange.signal_type.signal_base import SignalType
from threatexchange.signal_type.pdq.signal import PdqSignal
from threatexchange.signal_type.md5 import VideoMD5Signal
from threatexchange.signal_type.url import URLSignal
from threatexchange.signal_type.raw_text import RawTextSignal


@dataclass
class StopNCIICheckpoint(
    state.FetchCheckpointBase,
):
    update_time: int
    last_fetch_time: int

    def is_stale(self) -> bool:
        """Consider stale after 30d of not fetching"""
        return time.time() - self.last_fetch_time > 3600 * 24 * 30

    def get_progress_timestamp(self) -> t.Optional[int]:
        return self.update_time

    @classmethod
    def from_stopncii_fetch(
        cls, response: api.FetchHashesResponse
    ) -> "StopNCIICheckpoint":
        return cls(response.nextSetTimestamp, int(time.time()))


@dataclass
class StopNCIISignalMetadata(state.FetchedSignalMetadata):
    feedbacks: t.List[api.StopNCIICSPFeedback]

    def get_as_opinions(self) -> t.Sequence[state.SignalOpinion]:
        # TODO - handle which opinions are mine
        opinions = [
            state.SignalOpinion(False, _opinion_mapping(f.feedbackValue), f.tags)
            for f in self.feedbacks
        ]
        # implicitly, all records from StopNCII are from user-submitted cases
        opinions.append(
            state.SignalOpinion(
                False, state.SignalOpinionCategory.INVESTIGATION_SEED, set()
            ),
        )
        return opinions


@dataclass
class StopNCIICredentials(auth.CredentialHelper):
    ENV_VARIABLE: t.ClassVar[str] = "TX_STOPNCII_KEYS"
    FILE_NAME: t.ClassVar[str] = "~/.tx_stopncii_keys"

    fetch_function_key: str
    subscription_key: str
    base_url_override: t.Optional[str]

    @classmethod
    def _from_str(cls, s: str) -> "StopNCIICredentials":
        parts = s.strip().split(",")
        if len(parts) == 2:
            fetch_function_key, subscription_key = parts
            base_url_override = None
        elif len(parts) == 3:
            fetch_function_key, subscription_key, base_url_override = parts
        else:
            raise ValueError(
                "Invalid stopNCII credentials format. Should be 'function_key,subscription_key' OR 'function_key_subscription_key,base_url_override'."
            )

        return cls(
            subscription_key=subscription_key,
            fetch_function_key=fetch_function_key,
            base_url_override=base_url_override,
        )

    def _are_valid(self) -> bool:
        return bool(self.fetch_function_key and self.subscription_key)


class StopNCIISignalExchangeAPI(
    auth.SignalExchangeWithAuth[CollaborationConfigBase, StopNCIICredentials],
    signal_exchange_api.SignalExchangeAPIWithSimpleUpdates[
        CollaborationConfigBase,
        StopNCIICheckpoint,
        StopNCIISignalMetadata,
    ],
):
    """
    Conversion for the StopNCII.org API

    Key implementation details:
        1. Changes the key to be the SignalType names during fetch,
        2. Owner names are stored exposed as strings - no ids
        3. Both feedback and hash upload on API, but only feedback is available
           on server as of 4/14
    """

    def __init__(
        self,
        collab: CollaborationConfigBase,
        client: api.StopNCIIAPI,
    ) -> None:
        super().__init__()
        self.collab = collab
        self.api = client

    @staticmethod
    def get_config_cls() -> t.Type[CollaborationConfigBase]:
        return CollaborationConfigBase

    @staticmethod
    def get_checkpoint_cls() -> t.Type[StopNCIICheckpoint]:
        return StopNCIICheckpoint

    @staticmethod
    def get_record_cls() -> t.Type[StopNCIISignalMetadata]:
        return StopNCIISignalMetadata

    @staticmethod
    def get_credential_cls() -> t.Type[StopNCIICredentials]:
        return StopNCIICredentials

    @classmethod
    def for_collab(
        cls,
        collab: CollaborationConfigBase,
        credentials: t.Optional[StopNCIICredentials] = None,
    ) -> "StopNCIISignalExchangeAPI":
        credentials = credentials or StopNCIICredentials.get(cls)
        return cls(
            collab,
            api.StopNCIIAPI(
                credentials.subscription_key,
                credentials.fetch_function_key,
                base_url_override=credentials.base_url_override,
            ),
        )

    def fetch_iter(
        self,
        _supported_signal_types: t.Sequence[t.Type[SignalType]],
        checkpoint: t.Optional[StopNCIICheckpoint],
    ) -> t.Iterator[
        state.FetchDelta[
            t.Tuple[str, str],
            StopNCIISignalMetadata,
            StopNCIICheckpoint,
        ]
    ]:
        start_time = api.StopNCIIAPI.DEFAULT_START_TIME
        if checkpoint is not None:
            start_time = checkpoint.update_time
        for result in self.api.fetch_hashes_iter(start_timestamp=start_time):
            translated = (_get_delta_mapping(r) for r in result.hashRecords)
            yield state.FetchDelta(
                dict(t for t in translated if t[0][0]),
                StopNCIICheckpoint.from_stopncii_fetch(result),
            )


def _get_delta_mapping(
    record: api.StopNCIIHashRecord,
) -> t.Tuple[t.Tuple[str, str], t.Optional[StopNCIISignalMetadata]]:

    type_str = _type_mapping().get(record.signalType)
    if not type_str:
        return ("", ""), None

    # If no active cases associated with the hash, should it be deleted?
    if record.hashValue in (
        api.StopNCIICaseStatus.Withdrawn,
        api.StopNCIICaseStatus.Deleted,
    ):
        metadata = None
    else:
        metadata = StopNCIISignalMetadata(record.CSPFeedbacks)

    return ((type_str, record.hashValue), metadata)


@lru_cache(maxsize=1)
def _type_mapping() -> t.Dict[api.StopNCIISignalType, str]:
    return {
        api.StopNCIISignalType.ImagePDQ: PdqSignal.get_name(),
        api.StopNCIISignalType.VideoMD5: VideoMD5Signal.get_name(),
        api.StopNCIISignalType.URL: URLSignal.get_name(),
        api.StopNCIISignalType.Text: RawTextSignal.get_name(),
    }


def _opinion_mapping(fb: api.StopNCIICSPFeedbackValue) -> state.SignalOpinionCategory:
    if fb == api.StopNCIICSPFeedbackValue.Blocked:
        return state.SignalOpinionCategory.POSITIVE_CLASS
    if fb == api.StopNCIICSPFeedbackValue.NotBlocked:
        return state.SignalOpinionCategory.NEGATIVE_CLASS
    return state.SignalOpinionCategory.INVESTIGATION_SEED



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the pdf content type.
"""

from io import StringIO
import typing as t
from pathlib import Path

from threatexchange.content_type.content_base import ContentType
from threatexchange.content_type.text import TextContent

from pdfminer.converter import TextConverter
from pdfminer.layout import LAParams
from pdfminer.pdfdocument import PDFDocument
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.pdfpage import PDFPage
from pdfminer.pdfparser import PDFParser


class PDFContent(ContentType):
    """
    Content that in Portable Document Format.
    """

    @classmethod
    def extract_additional_content(
        cls, content_arg: str
    ) -> t.List[t.Tuple[t.Type[ContentType], str]]:
        path = Path(content_arg)
        if not path.is_file():
            raise Exception(f"Not a file: {content_arg}")
        if path.suffix != ".pdf":
            raise Exception(f"Not a .pdf: {content_arg}")

        text = StringIO()
        with path.open("rb") as in_file:
            parser = PDFParser(in_file)
            doc = PDFDocument(parser)
            rsrcmgr = PDFResourceManager()
            device = TextConverter(rsrcmgr, text, laparams=LAParams())
            interpreter = PDFPageInterpreter(rsrcmgr, device)
            for page in PDFPage.create_pages(doc):
                interpreter.process_page(page)
        return [
            (TextContent, text.getvalue()),
        ]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import unittest

from threatexchange.signal_type.url_md5 import UrlMD5Signal

URL_TEST = "www.facebook.com/?user=123"
FULL_URL_TEST = "https://www.facebook.com/?user=123"
URL_TEST_MD5 = "e359430911fe80c2dd526d3cca21da30"


class UrlMD5SignalTestCase(unittest.TestCase):
    def test_can_hash_simple_url(self):
        assert URL_TEST_MD5 == UrlMD5Signal.hash_from_str(
            URL_TEST
        ), "MD5 hash does not match"

    def test_can_hash_full_url(self):
        assert URL_TEST_MD5 == UrlMD5Signal.hash_from_str(
            FULL_URL_TEST
        ), "MD5 hash does not match"



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Abstraction for different content types.

This records all the valid signal types for a piece of content.
"""

import typing as t

from threatexchange import common


class ContentType:
    @classmethod
    def get_name(cls) -> str:
        """The human-friendly display name"""
        return common.class_name_to_human_name(cls.__name__, "Content")

    @classmethod
    def extract_additional_content(
        cls, content_arg: str
    ) -> t.List[t.Tuple[t.Type["ContentType"], str]]:
        """
        Post-process/download content to find additional components

        Examples might be:
        * URL => download page content and return photo/video/text/URL snippets
        * File => Identify content based on file type and return appropriate snippets
        * Photo => run OCR and extract text
        * Video => break out photo thumbnail, close caption text, audio
        """
        return []



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Helpers and wrappers around the /threat_updates endpoint.
"""

import json
import os
import pathlib
import time
import typing as t
from dataclasses import dataclass

from .api import ThreatExchangeAPI, _CursoredResponse
from .descriptor import SimpleDescriptorRollup


class ThreatUpdateSerialization:
    """
    A wrapper for converting records fetched from /threat_updates
    """

    @property
    def key(self):
        """This should either be the indicator type+string or id"""
        raise NotImplementedError

    @classmethod
    def from_threat_updates_json(cls, app_id: int, te_json):
        raise NotImplementedError

    @classmethod
    def te_threat_updates_fields(cls):
        """Which &fields= arguments need to be passed for this serialization"""
        raise NotImplementedError


@dataclass
class ThreatUpdateJSON(ThreatUpdateSerialization):
    """A thin wrapper around the /threat_updates API return"""

    raw_json: t.Dict[str, t.Any]

    @property
    def should_delete(self) -> bool:
        """This record is a tombstone, and we should delete our copy"""
        # This should just be should_delete only, but see
        # https://github.com/facebook/ThreatExchange/issues/834
        return self.raw_json["should_delete"] or "descriptors" not in self.raw_json

    @property
    def key(self):
        return self.id

    @property
    def id(self) -> int:
        return int(self.raw_json["id"])

    @property
    def indicator(self) -> str:
        return self.raw_json["indicator"]

    @property
    def threat_type(self) -> str:
        return self.raw_json["type"]

    @property
    def time(self) -> int:
        """The time of the update"""
        return int(self.raw_json["last_updated"])

    @classmethod
    def from_threat_updates_json(cls, app_id, te_json):
        return cls(te_json)

    @classmethod
    def te_threat_updates_fields(cls) -> t.Tuple[str, ...]:
        # Could also return empty here, but this set is useful for basically
        # any serialization, and so it makes sense to fetch it for future processing
        # (even though it's verbose)
        return SimpleDescriptorRollup.te_threat_updates_fields()


class ThreatUpdatesDelta:
    """
    A class for tracking a raw stream of /threat_updates

    Any integration with ThreatExchange involves the creation of a local copy
    of the data. /threat_updates sends changes to that data in either the form
    of an insert/update or a delete.

    A delta is a stream of updates and deletes, which when applied to an
    existing database will give you the current set of live records.

    As a parallelization trick, if you need to fetch between t1 and t3,
    you can pick a point between them, t2, and fetch [t1, t2) and [t2, t3)
    simulatenously, and the merging of the two is guaranteed to be the same
    as [t1, t3). The split() and merge() commands aid with this operation
    """

    def __init__(
        self,
        privacy_group: int,
        start: int = 0,
        end: t.Optional[int] = None,
        types: t.Iterable[str] = (),
    ) -> None:
        self.privacy_group = privacy_group
        self.updates: t.List = []
        self.current = start
        self.start = start
        self.end = end
        self.types = list(types)

        self._cursor: t.Optional[_CursoredResponse] = None

    @property
    def done(self) -> bool:
        """Has this delta fetched its entire assigned range?"""
        return bool(self.end and self.end <= self.current)

    def __bool__(self):
        return self.done or bool(self.updates)

    def __iter__(self):
        return iter(self.updates)

    def merge(self, delta: "ThreatUpdatesDelta") -> None:
        """
        Merge the earlier delta (this object) with a later delta.

        If you have
        t1 ---> t2 ---> t3
        t1.merge(t2).merge(t3) is valid, and will give you a range from t1-t3
        """
        if not self.done or self.end != delta.start:
            raise ValueError("unchecked merge!")
        self.updates.extend(delta.updates)
        self.current = delta.current
        self.end = delta.end

    def one_fetch(self, api: ThreatExchangeAPI):
        """
        Do a single fetch from ThreatExchange and store the results.

        One fetch only, please.
        """
        if self.done:
            return
        now = time.time()
        if not self._cursor:
            self._cursor = api.get_threat_updates(
                self.privacy_group,
                page_size=500,
                start_time=self.start,
                stop_time=self.end,
                types=self.types,
                fields=ThreatUpdateJSON.te_threat_updates_fields(),
                decode_fn=ThreatUpdateJSON,
            )
        for update in self._cursor.next():
            self.updates.append(ThreatUpdateJSON(update.raw_json))
            # Is supposed to be strictly increasing
            self.current = max(update.time, self.current)
        if self._cursor.done:
            if not self.end:
                self.end = int(now)
            self.current = self.end

        return self._cursor.data

    def split(
        self, n: int
    ) -> t.Tuple["ThreatUpdatesDelta", t.List["ThreatUpdatesDelta"]]:
        """Split this delta into n deltas of roughly even size"""
        tar = self.end or time.time()
        diff = int((tar - self.current) // (n + 1))
        if diff <= 0:
            return self, []
        end = self.end
        prev = self
        new_deltas = []
        for i in range(n - 1):
            new_start = prev.start + diff
            new_deltas.append(
                ThreatUpdatesDelta(
                    self.privacy_group,
                    new_start,
                )
            )
        prev.end = end

        return self, new_deltas

    def incremental_sync_from_threatexchange(
        self,
        api: ThreatExchangeAPI,
        *,
        limit: t.Optional[int] = None,
        progress_fn=lambda x: None,
    ) -> None:
        """
        Fetch from threat_updates to get a more up-to-date copy of the data.
        """

        # TODO actually implement fancy threading logic
        # alternative - instead make the API give hints about where to start
        # fetches, which will mean that fancy threading logic will be simpler
        while not self.done:
            for update in self.one_fetch(api):
                progress_fn(update)
                if limit is not None:
                    limit -= 1
                    if limit <= 0:
                        return


class ThreatUpdateCheckpoint(t.NamedTuple):
    """
    State about the progress of a /threat_updates-backed state.

    If a client does not resume tailing the threat_updates endpoint fast enough,
    deletion records will be removed, making it impossible to determine which
    records should be retained without refetching the entire dataset from scratch.
    The API implementation will retain for 90 days:
    https://developers.facebook.com/docs/threat-exchange/reference/apis/threat-updates/
    """

    # See docstring about tailing fast enough
    DEFAULT_REFETCH_SEC: int = 3600 * 24 * 85  # 85 days

    # When was the last time we started or the furthest we've seen,
    # to check against the store getting too stale
    last_fetch_time: int = 0
    # Where should we resume from?
    fetch_checkpoint: int = 0

    def get_updated(self, delta: ThreatUpdatesDelta) -> "ThreatUpdateCheckpoint":
        # If starting from 0, this is the first fetch, in which case the first update
        # means that we fetched now.
        last_fetch_time = self.last_fetch_time
        if last_fetch_time == 0:
            last_fetch_time = int(time.time())

        return ThreatUpdateCheckpoint(
            last_fetch_time=max(last_fetch_time, delta.current),
            fetch_checkpoint=max(self.fetch_checkpoint, delta.current),
        )

    @property
    def stale(self):
        """Is this checkpoint so old as to be invalid?"""
        return self.last_fetch_time + self.DEFAULT_REFETCH_SEC < time.time()


class ThreatUpdatesStore:
    """
    A wrapper for ThreatIndicator records for a single Collaboration

    There is a unique file for each combination of:
     * IndicatorType
     * PrivacyGroup

    The contents of file does not strip anything from the API
    response, so can potentially contain a lot of data.
    """

    def __init__(
        self,
        privacy_group: int,
    ) -> None:
        self.privacy_group = privacy_group
        self.checkpoint: t.Optional[ThreatUpdateCheckpoint] = None

    @property
    def fetch_checkpoint(self):
        return self.checkpoint.fetch_checkpoint

    def reset(self) -> None:
        """Toss old state and begin anew"""
        self.checkpoint = ThreatUpdateCheckpoint()

    @property
    def next_delta(self) -> ThreatUpdatesDelta:
        """Return the next delta that should be applied"""
        return ThreatUpdatesDelta(
            self.privacy_group,
            self.checkpoint.fetch_checkpoint if self.checkpoint else 0,
            None,
        )

    def load_checkpoint(self) -> None:
        self.checkpoint = self._load_checkpoint()

    def _load_checkpoint(self) -> ThreatUpdateCheckpoint:
        """Load the state of the threat_updates checkpoints"""
        raise NotImplementedError

    def _store_checkpoint(self, checkpoint: ThreatUpdateCheckpoint) -> None:
        """Save the state of the threat_updates checkpoints after a succesful apply"""
        raise NotImplementedError

    def _apply_updates_impl(
        self,
        delta: ThreatUpdatesDelta,
        post_apply_fn=lambda x: None,
    ) -> None:
        """Apply delta to state and store it"""
        raise NotImplementedError

    @property
    def stale(self) -> bool:
        """Is this state so old that it might be invalid?"""
        return self.checkpoint.stale if self.checkpoint else False

    def apply_updates(
        self,
        delta: ThreatUpdatesDelta,
        post_apply_fn=lambda x: None,
    ) -> None:
        """Merge updates to the data store"""
        if delta.start != 0:
            assert (
                self.checkpoint and delta.start <= self.checkpoint.fetch_checkpoint
            ), "gap in delta record"
            assert not self.stale, "attempted to apply stale delta"

        # It's possible the fetch completed but has no records
        if delta.updates:
            self._apply_updates_impl(delta, post_apply_fn)
        if self.checkpoint:
            self.checkpoint = self.checkpoint.get_updated(delta)
            self._store_checkpoint(self.checkpoint)


class ThreatUpdateFileStore(ThreatUpdatesStore):
    """
    A simple file storage (in lieu of DB) with in-memory merge
    """

    def __init__(
        self,
        state_dir: pathlib.Path,
        privacy_group: int,
        app_id: int,
        *,
        serialization=ThreatUpdateJSON,
    ) -> None:
        super().__init__(privacy_group)
        self.path = state_dir
        self.app_id = app_id
        self._serialization = serialization
        self._cached_state: t.Optional[t.Dict[str, t.Any]] = None

    @property
    def checkpoint_file(self) -> pathlib.Path:
        return self.path / f"{self.privacy_group}.threat_updates.checkpoint"

    def reset(self):
        super().reset()
        if self._cached_state:
            self._cached_state.clear()

    def _load_checkpoint(self) -> ThreatUpdateCheckpoint:
        """Load the state of the threat_updates checkpoints from state directory"""
        if not self.checkpoint_file.exists():
            return ThreatUpdateCheckpoint()

        with self.checkpoint_file.open("r") as f:
            checkpoint_json = json.load(f)
            return ThreatUpdateCheckpoint(
                checkpoint_json["last_fetch_time"],
                checkpoint_json["fetch_checkpoint"],
            )

    def _store_checkpoint(self, checkpoint: ThreatUpdateCheckpoint) -> None:
        with self.checkpoint_file.open("w") as f:
            json.dump(
                {
                    "last_fetch_time": checkpoint.last_fetch_time,
                    "fetch_checkpoint": checkpoint.fetch_checkpoint,
                },
                f,
                indent=2,
            )

    def load_state(self, allow_cached=True):
        if not self.path.exists():
            return {}
        if not allow_cached or self._cached_state is None:
            self._cached_state = {
                item.key: item for item in self._serialization.load(self.path)
            }
        return self._cached_state

    def _apply_updates_impl(
        self, delta: ThreatUpdatesDelta, post_apply_fn=lambda x: None
    ) -> None:
        os.makedirs(self.path, exist_ok=True)
        state = {}
        if delta.start > 0:
            state = self.load_state()
        for update in delta:
            item = self._serialization.from_threat_updates_json(
                self.app_id, update.raw_json
            )
            if update.should_delete:
                state.pop(item.key, None)
            else:
                state[item.key] = item

        self._cached_state = state
        self._serialization.store(self.path, state.values())



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import os
import unittest
import collections.abc

from threatexchange.exchanges.clients.fb_threatexchange.api import ThreatExchangeAPI
from threatexchange.exchanges.clients.fb_threatexchange.api_representations import (
    ThreatPrivacyGroup,
)

THREAT_EXCHANGE_INTEGRATION_TEST_TOKEN = os.getenv(
    "THREAT_EXCHANGE_INTEGRATION_TEST_TOKEN"
)


@unittest.skipUnless(
    THREAT_EXCHANGE_INTEGRATION_TEST_TOKEN,
    "Integration Test requires tokens. Use THREAT_EXCHANGE_INTEGRATION_TEST_TOKEN environment variable.",
)
class APIIntegrationTest(unittest.TestCase):
    def setUp(self):
        self.api = ThreatExchangeAPI(THREAT_EXCHANGE_INTEGRATION_TEST_TOKEN)

    def test_get_threat_privacy_groups_member(self):
        """
        Assumes that the app (if token is provided) will have at least one
        privacy group.
        """
        response = self.api.get_threat_privacy_groups_member()
        self.assertTrue(
            isinstance(response, collections.abc.Sequence)
            and not isinstance(response, staticmethod),
            "API returned something that's not a list!",
        )

        self.assertTrue(isinstance(response[0], ThreatPrivacyGroup))



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the Trend Query (keywords and regexes) content type.
"""

import json
import re
import typing as t
from threatexchange.content_type.content_base import ContentType
from threatexchange.content_type.text import TextContent

from threatexchange.signal_type import signal_base
from threatexchange.signal_type import index
from threatexchange.exchanges.impl.fb_threatexchange_signal import (
    HasFbThreatExchangeIndicatorType,
)


class TrendQuery:
    """
    A parsed trend query, based on regexes.
    """

    REGEX_PREFIX = "regex-"

    # re.pattern is not in 3.6, which is what we are targeting right now
    def __init__(self, query_json: t.Dict[str, t.Any]) -> None:
        self.and_terms: t.List[t.List[t.Any]] = [
            [self._parse_term(t) for t in and_["or"]] for and_ in query_json["and"]
        ]
        self.not_terms: t.List[t.Any] = [self._parse_term(t) for t in query_json["not"]]

    def _parse_term(self, t) -> t.Any:
        if t.startswith(self.REGEX_PREFIX):
            return re.compile(t[len(self.REGEX_PREFIX) + 1 : -1])
        return re.compile(f"\\b{re.escape(t)}\\b")

    def _match_term(self, t: t.Union[str, t.Any], text: str) -> bool:
        return bool()

    def matches(self, text: str) -> bool:
        for or_ in self.and_terms:
            if not any(t.search(text) for t in or_):
                break
        else:
            return not any(t.search(text) for t in self.not_terms)
        return False


class TrendQuerySignal(
    signal_base.SignalType, signal_base.MatchesStr, HasFbThreatExchangeIndicatorType
):
    """
    Trend Queries are a combination of and/or/not regexes.

    Based off a system effective for grouping content at Facebook, Trend Queries
    can potentially help you sift though large sets of content to quickly flag
    ones that might be interesting to you.

    They have high "recall" but potentially low "precision".
    """

    INDICATOR_TYPE = "TREND_QUERY"

    @classmethod
    def get_content_types(cls) -> t.List[t.Type[ContentType]]:
        return [TextContent]

    @classmethod
    def validate_signal_str(cls, signal_str: str) -> str:
        tq = TrendQuery(
            json.loads(signal_str)
        )  # TODO - does this throw all the right exceptions?
        return signal_str

    @classmethod
    def matches_str(
        cls, hash: str, haystack: str
    ) -> signal_base.SignalComparisonResult:
        tq = TrendQuery(json.loads(hash))
        return signal_base.SignalComparisonResult.from_bool_only(tq.matches(haystack))

    @classmethod
    def get_index_cls(cls) -> t.Type[index.SignalTypeIndex]:
        return TrendQueryIndex

    @staticmethod
    def get_examples() -> t.List[str]:
        return [
            json.dumps(
                {
                    "and": [
                        {
                            "or": [
                                "basketball",
                                "basket ball",
                                "basket-ball",
                                "bball",
                                "hoops",
                            ]
                        },
                        {"or": ["play", "tonight", "today", "now"]},
                    ],
                    "not": ["tomorrow", "baseball", "hockey", "football", "soccer"],
                }
            )
        ]


class TrendQueryIndex(index.SignalTypeIndex[index.T]):
    def __init__(self) -> None:
        self.state: t.Dict[str, t.Tuple[TrendQuery, t.List[index.T]]] = {}

    # TODO - Figure out how to properly capture hash vs search
    def query(self, hash: str) -> t.List[index.IndexMatch[index.T]]:
        ret: t.List[index.IndexMatch[index.T]] = []
        for tq, values in self.state.values():
            if tq.matches(hash):
                ret.extend(
                    index.IndexMatch(index.SignalSimilarityInfo(), v) for v in values
                )
        return ret

    def add(self, hash: str, value: index.T) -> None:
        old_val = self.state.get(hash)
        query_json = json.loads(hash)

        if old_val is None:
            self.state[hash] = (
                TrendQuery(query_json),
                [value],
            )
        else:
            old_val[1].append(value)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import unittest
import pathlib

from threatexchange.signal_type.md5 import VideoMD5Signal

TEST_FILE = pathlib.Path(__file__).parent.parent.parent.parent.joinpath(
    "data", "sample-b.jpg"
)


class VideoMD5SignalTestCase(unittest.TestCase):
    def setUp(self):
        self.a_file = open(TEST_FILE, "rb")

    def tearDown(self):
        self.a_file.close()

    def test_can_hash_simple_files(self):
        assert "d35c785545392755e7e4164457657269" == VideoMD5Signal.hash_from_bytes(
            self.a_file.read()
        ), "MD5 hash does not match"



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Core abstractions for signal types.
"""

import abc
import pathlib
import typing as t

from threatexchange import common
from threatexchange.content_type import content_base
from threatexchange.signal_type import index


class SignalComparisonResult(t.NamedTuple):
    match: bool
    distance: index.SignalSimilarityInfo

    @classmethod
    def from_bool_only(cls, matches: bool) -> "SignalComparisonResult":
        """For SignalTypes with no distance"""
        return cls.from_match_only() if matches else cls.from_no_match_only()

    @classmethod
    def from_match_only(cls) -> "SignalComparisonResult":
        """For SignalTypes with no distance"""
        return cls(True, index.SignalSimilarityInfo())

    @classmethod
    def from_no_match_only(cls) -> "SignalComparisonResult":
        """For SignalTypes with no distance"""
        return cls(False, index.SignalSimilarityInfo())

    @classmethod
    def from_simple_dist(
        cls, dist: index.CT, threshold: index.CT
    ) -> "SignalComparisonResult":
        """For SignalTypes with simple distance"""
        return cls(
            dist <= threshold,
            index.SignalSimilarityInfoWithSingleDistance[index.CT](dist),
        )

    @classmethod
    def from_dist(
        cls, dist: index.SignalSimilarityInfo, threshold: index.SignalSimilarityInfo
    ) -> "SignalComparisonResult":
        """For SignalTypes w"""
        return cls(
            dist <= threshold,
            dist,
        )


class SignalType(abc.ABC):
    """
    Abstraction for different signal types.

    A signal type is an intermediate representation of content that can be used
    to match against similar or identical content. Sometimes called a "hash"
    type.

    This class additionally helps translates ThreatDescriptors into the correct
    representation to do matching, as well as serialize that representation
    into a compact form.

    # Why is `signal_str` str?
    Forcing the type to always be serializable in str gives us the advantage that
    uploading it to SignalAPIs is always straightforward.
    """

    @classmethod
    def get_name(cls) -> str:
        """A compact name in lower_with_underscore style (used in filenames)"""
        return common.class_name_to_human_name(cls.__name__, "Signal")

    @classmethod
    @abc.abstractmethod
    def get_content_types(cls) -> t.List[t.Type[content_base.ContentType]]:
        """Which content types this Signal applies to (usually just one)"""
        pass

    @classmethod
    @abc.abstractmethod
    def get_index_cls(cls) -> t.Type[index.SignalTypeIndex]:
        """Return the index class that handles this signal type"""
        # Confused about which one to start with?
        # Make a subclass of:
        #   1. TrivialLinearSearchHashIndex: Just use compare_hash
        #   2. TrivialLinearSearchMatchIndex: If you are MatchesStr
        # Or if you only support exact matching (like MD5):
        #   3. TrivialSignalTypeIndex
        pass

    @classmethod
    def validate_signal_str(cls, signal_str: str) -> str:
        """
        Return a normalized version the signal or throws an exception if malformed
        """
        if not signal_str:
            raise ValueError("empty hash")
        return signal_str.strip()

    @staticmethod
    @abc.abstractmethod
    def get_examples() -> t.List[str]:
        """
        Return example signals, which can be used for tests or demos.

        Strings returned by this class are automatically pulled in by the
        StaticSampleSignalExchangeAPI, which it will "fetch" and build
        indices for if the Signal is enabled.
        """
        return []


class FileHasher(abc.ABC):
    """
    This class can hash files.
    """

    @classmethod
    @abc.abstractmethod
    def compare_hash(cls, hash1: str, hash2: str) -> SignalComparisonResult:
        """
        Compare the distance of two hashes, the key operation for matching.

        Note that this can just be a reference/helper, and the efficient
        version of the algorithm can live in the index class.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def hash_from_file(cls, file: pathlib.Path) -> str:
        """
        Get a string representation of the hash from a file.

        If a hash cannot be generated, empty string should be returned.
        """
        pass


class TextHasher(FileHasher):
    """
    This class can turn text into intermediary representations (hashes)
    """

    @classmethod
    @abc.abstractmethod
    def hash_from_str(cls, text: str) -> str:
        """
        Get a string representation of the hash from a string.

        If a hash cannot be generated, empty string should be returned.
        """
        pass

    @classmethod
    def hash_from_file(cls, file: pathlib.Path) -> str:
        return cls.hash_from_str(file.read_text())


class MatchesStr(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def matches_str(cls, signal: str, haystack: str) -> SignalComparisonResult:
        """
        Compare the signal and text, the key operation for matching.

        Note that this can just be a reference/helper, and the efficient
        version of the algorithm can live in the index class.
        """
        pass


class BytesHasher(FileHasher):
    """
    This class can hash bytes.
    """

    @classmethod
    @abc.abstractmethod
    def hash_from_bytes(cls, bytes_: bytes) -> str:
        """
        Get a string representation of the hash from bytes.

        If a hash cannot be generated, empty string should be returned.
        """
        pass

    @classmethod
    def hash_from_file(cls, file: pathlib.Path) -> str:
        return cls.hash_from_bytes(file.read_bytes())


class SimpleSignalType(SignalType):
    """
    Dead simple implementation for loading/storing a SignalType.

    Assumes that the signal type can easily merge on a string.
    """

    @classmethod
    def compare_hash(cls, hash1: str, hash2: str) -> SignalComparisonResult:
        return SignalComparisonResult.from_bool_only(hash1 == hash2)


class TrivialSignalTypeIndex(index.SignalTypeIndex[index.T]):
    """
    Index that does only exact matches
    """

    def __init__(self) -> None:
        self.state: t.Dict[str, t.List[index.T]] = {}

    def query(self, query: str) -> t.List[index.IndexMatch[index.T]]:
        return [
            index.IndexMatch(index.SignalSimilarityInfo(), meta)
            for meta in self.state.get(query, [])
        ]

    def add(self, signal_str: str, entry: index.T) -> None:
        l = self.state.get(signal_str)
        if not l:
            l = []
            self.state[signal_str] = l
        l.append(entry)


class TrivialLinearSearchHashIndex(index.SignalTypeIndex[index.T]):
    """
    Index that does a linear search and serializes with pickle

    O(n) is the best n, clearly.
    """

    # You'll have to override with each usecase, because I wasn't sure
    # If pickle would behave expectedly here
    _SIGNAL_TYPE: t.Type[FileHasher]

    def __init__(self) -> None:
        self.state: t.List[t.Tuple[str, index.T]] = []

    def query(self, query_hash: str) -> t.List[index.IndexMatch[index.T]]:
        ret = []
        for hash, payload in self.state:
            res = self._SIGNAL_TYPE.compare_hash(hash, query_hash)
            if res.match:
                ret.append(index.IndexMatch(res.distance, payload))
        return ret

    def add(self, signal_str: str, entry: index.T) -> None:
        self.state.append((signal_str, entry))


class TrivialLinearSearchMatchIndex(index.SignalTypeIndex[index.T]):
    """
    Index that does a linear search and serializes with pickle

    O(n) is the best n, clearly.
    """

    # You'll have to override with each usecase
    _SIGNAL_TYPE: t.Type[MatchesStr]

    def __init__(self) -> None:
        self.state: t.List[t.Tuple[str, index.T]] = []
        assert issubclass(self._SIGNAL_TYPE, MatchesStr)

    def query(self, query_hash: str) -> t.List[index.IndexMatch[index.T]]:
        ret = []
        for signal, payload in self.state:
            res = self._SIGNAL_TYPE.matches_str(signal, query_hash)
            if res.match:
                ret.append(index.IndexMatch(res.distance, payload))
        return ret

    def add(self, signal_str: str, entry: index.T) -> None:
        self.state.append((signal_str, entry))



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Base classes for passing data between SignalExchangeAPIs and other interfaces.

Many implementations will choose to extend these to add additional metadata
needed to power API features.
"""


from dataclasses import dataclass
from enum import IntEnum
from abc import ABC, abstractmethod
import typing as t

from threatexchange.exchanges.collab_config import CollaborationConfigBase
from threatexchange.signal_type.signal_base import SignalType

Self = t.TypeVar("Self")

# Note - key is artificially restricted, but it's simple to add more or even
#        remove the restriction entirely if eventually needed
TUpdateRecordKey = t.TypeVar("TUpdateRecordKey", str, int, t.Tuple[str, str])
TUpdateRecordValue = t.TypeVar("TUpdateRecordValue")


@dataclass
class FetchCheckpointBase:
    """
    If you need to store checkpoint information, this is the place to do it
    """

    def is_stale(self) -> bool:
        """
        For some APIs, stored state may become invalid if stored too long.

        Return true if the old data should be deleted and fetched from scratch.
        """
        return False  # Default, assume checkpoints never expire

    def get_progress_timestamp(self) -> t.Optional[int]:
        """If the checkpoint can, give the time it corresponds to"""
        return None


class NoCheckpointing(FetchCheckpointBase):
    """
    A checkpoint for SignalExchangeAPIs that always fetch the entire state
    """

    def is_stale(self) -> bool:
        return True


TFetchCheckpoint = t.TypeVar(
    "TFetchCheckpoint", covariant=True, bound=FetchCheckpointBase
)


@dataclass
class FetchDelta(t.Generic[TUpdateRecordKey, TUpdateRecordValue, TFetchCheckpoint]):
    """
    Represents an incremental chunk of state fetched from an API.

    The exact storage format may vary by API, and a naive implementation
    for combining updates to an in-memory state lives in the API, but it
    often makes sense for a dedicated storage implementation to have a
    more efficient implementation.

    @see FetchStore
    """

    updates: t.Dict[TUpdateRecordKey, t.Optional[TUpdateRecordValue]]
    checkpoint: TFetchCheckpoint


FetchDeltaTyped = FetchDelta[
    t.Any, t.Any, FetchCheckpointBase
]  # to anchor checkpoint for typing


class SignalOpinionCategory(IntEnum):
    """
    What the opinion on a signal is.

    Some APIs may not support all of these, but each of these should influence
    what action you might take as a result of matching, otherwise it might
    make more sense as a tag.
    """

    # Signal generates false positives, or is test/example data
    NEGATIVE_CLASS = 0
    # Indirect indicator, requires verifying. Can lead to POSITIVE_CLASS content
    INVESTIGATION_SEED = 1
    # Signal corresponds to content that fits collaboration categories
    POSITIVE_CLASS = 2


@dataclass
class SignalOpinion:
    """
    The metadata of a single signal upload.

    Certain APIs won't have any concept of owner, category, or tags,
    in which case get_trivial() is reasonable
    default.

    Some implementations may extend this to store additional API-specific data

    @see threatexchange.fetch_api.SignalExchangeAPI
    """

    # This opinion was generated by me, rather than an external entity
    is_mine: bool
    category: SignalOpinionCategory
    tags: t.Set[str]

    @classmethod
    def get_trivial(cls):
        return cls(False, SignalOpinionCategory.INVESTIGATION_SEED, [])

    def __setstate__(self, d: t.Dict[str, t.Any]) -> None:
        """Implemented for pickle version compatibility."""
        # 0.99.0 => 1.0.0:
        ### field 'owner': int replaced to 'is_mine': bool
        if "owner" in d:
            d.pop("owner")
            d["is_mine"] = False
        self.__dict__ = d


class AggregateSignalOpinionCategory(IntEnum):
    """
    Represent multiple opinions as one.

    Keep in Sync with SignalOpinionCategory
    """

    NEGATIVE_CLASS = 0
    INVESTIGATION_SEED = 1
    POSITIVE_CLASS = 2
    DISPUTED = 3  # Some positive, some negative

    @classmethod
    def from_opinion_categories(
        cls, opinion_categories: t.Iterable[SignalOpinionCategory]
    ) -> "AggregateSignalOpinionCategory":
        aggregate_opinion = None
        for category in opinion_categories:
            aggregate_opinion = cls.aggregate(aggregate_opinion, category)
        assert aggregate_opinion is not None
        return aggregate_opinion

    @classmethod
    def aggregate(
        cls,
        old: t.Optional["AggregateSignalOpinionCategory"],
        new: t.Union["AggregateSignalOpinionCategory", SignalOpinionCategory],
    ) -> "AggregateSignalOpinionCategory":
        """
        Combine signal opinions into an aggregate opinion.

        In general, take the highest confidence/severity of true positives,
        unless you have both a true + false positive, in which case the result
        is disputed.
        """
        new = AggregateSignalOpinionCategory(new)
        if old is None:
            return new
        lo = min(old, new)
        hi = max(old, new)
        if lo == hi:
            return hi
        return cls.DISPUTED if lo == cls.NEGATIVE_CLASS else hi


@dataclass
class AggregateSignalOpinion:

    category: AggregateSignalOpinionCategory
    tags: t.Set[str]

    @classmethod
    def from_opinions(
        cls, opinions: t.Sequence[SignalOpinion]
    ) -> "AggregateSignalOpinion":
        assert opinions
        return cls(
            tags={t for o in opinions for t in o.tags},
            category=AggregateSignalOpinionCategory.from_opinion_categories(
                o.category for o in opinions
            ),
        )


@dataclass
class FetchedSignalMetadata:
    """
    Metadata to make decisions on matches and power feedback on the fetch API.

    You likely need to extend this for your API to include enough context for
    SignalExchangeAPI.report_seen() and others.

    If your API supports multiple databases or collections, you likely
    will need to store that here.
    """

    def get_as_opinions(self) -> t.Sequence[SignalOpinion]:
        return [SignalOpinion.get_trivial()]

    def get_as_aggregate_opinion(self) -> AggregateSignalOpinion:
        return AggregateSignalOpinion.from_opinions(self.get_as_opinions())

    def __str__(self) -> str:
        """
        A human-readable version of the opinion suitable for the terminal
        """
        agg = self.get_as_aggregate_opinion()
        if not agg.tags:
            return agg.category.name
        return f"{agg.category.name} {','.join(agg.tags)}"


TFetchedSignalMetadata = t.TypeVar(
    "TFetchedSignalMetadata", covariant=True, bound=FetchedSignalMetadata
)


# TODO t.Generic[TFetchDeltaBase, TFetchedSignalDataBase, FetchCheckpointBase]
#      to help keep track of the expected subclasses for an impl
class FetchedStateStoreBase(ABC):
    """
    An interface to previously fetched or persisted state.

    You will need to extend this for your API, but even worse, there
    might need to be multiple versions for a single API if it's being
    used by Hasher-Matcher-Actioner, since that might want to specialcase
    for AWS components.

    = A Note on Metadata ID =
    It's assumed that the storage will be split into a scheme that allows
    addressing individual IDs. Depending on the implementation, you may
    have to invent IDs during merge() which will also need to be persisted,
    since they need to be consistent between instanciation
    """

    @abstractmethod
    def get_checkpoint(
        self, collab: CollaborationConfigBase
    ) -> t.Optional[FetchCheckpointBase]:
        """
        Returns the last checkpoint passed to merge() after a flush()
        """
        raise NotImplementedError

    @abstractmethod
    def merge(self, collab: CollaborationConfigBase, delta: FetchDelta) -> None:
        """
        Merge a FetchDelta into the state.

        At the implementation's discretion, it may call flush() or the
        equivalent work.
        """
        raise NotImplementedError

    @abstractmethod
    def flush(self) -> None:
        """
        Finish writing the results of previous merges to persistant state.

        This should also persist the checkpoint.
        """
        raise NotImplementedError

    @abstractmethod
    def clear(self, collab: CollaborationConfigBase) -> None:
        """
        Delete all the stored state for this collaboration.
        """
        raise NotImplementedError

    @abstractmethod
    def get_for_signal_type(
        self, collabs: t.List[CollaborationConfigBase], signal_type: t.Type[SignalType]
    ) -> t.Dict[str, t.Dict[str, FetchedSignalMetadata]]:
        """
        Get as a map of CollabConfigBase.name() => {signal: Metadata}

        This is meant for simple storage and indexing solutions, but at
        scale, you likely want to store as IDs rather than the full metadata.

        TODO: This currently implies that you are going to load the entire dataset
        into memory, which once we start getting huge amounts of data, might not make
        sense.
        """
        raise NotImplementedError



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import random
from functools import reduce
from threatexchange.signal_type.pdq.pdq_utils import PDQ_HEX_STR_LEN, BITS_IN_PDQ


def get_zero_hash() -> str:
    """Return a pdq hash hex str that is zero for every byte"""
    return "0" * PDQ_HEX_STR_LEN


def get_random_hash() -> str:
    """Return a random pdq hash hex str"""
    return f"{random.randrange(2**256):0{PDQ_HEX_STR_LEN}x}"


def get_similar_hash(pdq_hex: str, dist: int) -> str:
    """Return a pdq hash hex str with dist hamming distance away from pdq_hex"""
    if not (0 <= dist <= BITS_IN_PDQ):
        raise ValueError("Invalid distance value")
    order = random.sample(range(BITS_IN_PDQ), k=dist)
    pdq_int = reduce(lambda x, y: x ^ (1 << y), order, int(pdq_hex, 16))
    return f"{pdq_int:0{PDQ_HEX_STR_LEN}x}"



import pytest
import typing as t
from threatexchange.exchanges.clients.stopncii.api import (
    StopNCIIAPI,
    StopNCIICaseStatus,
    StopNCIICSPFeedbackValue,
    StopNCIIEndpoint,
    StopNCIIHashRecord,
    StopNCIISignalType,
    StopNCIICSPFeedback,
)

PAGE_TOKEN = (
    "W3siY29tcG9zaXRlVG9rZW4iOnsidG9rZW4iOiIrUklEOn4xMnRWQUo4SzU2Y0NB"
    "QUFBQUFBQUFBPT0jUlQ6MSNUUkM6MiNSVEQ6SFBKSm1LVDF1WnFsQkJuR3lOU09C"
    "TUhaRytHL1FBPT0jSVNWOjIjSUVPOjY1NTY3"
)


def mock_get_impl(endpoint: str, **json):
    assert endpoint == StopNCIIEndpoint.FetchHashes

    if json.get("nextPageToken") != PAGE_TOKEN:
        return {
            "count": 2,
            "nextSetTimestamp": 1625175071,
            "nextPageToken": PAGE_TOKEN,
            "hasMoreRecords": True,
            "hashRecords": [
                {
                    "lastModtimestamp": 1625167804,
                    "hashValue": "2afc4a5c09628a7961c14d436493bba66b89b831453baa1d556ba385554daa82",
                    "hashStatus": "Received",
                    "caseNumbers": {
                        "27664732-76e1-4a17-8099-455798e67022": "Received",
                        "a2696edf-1237-4eb8-a9c1-cd8ee6c055e5": "Received",
                        "9fbb73d7-c177-4ed2-b4e8-0050b72decd0": "Received",
                        "bd2bff7c-8160-4615-8dce-38c2f066d1ac": "Received",
                    },
                    "signalType": "ImagePDQ",
                },
                {
                    "lastModtimestamp": 1625167844,
                    "hashValue": "79e07de27d7295339435d63cd31cf35a7bfa29eb2885008500a588a5ea3ae75a",
                    "hashStatus": "Received",
                    "caseNumbers": {"bd2bff7c-8160-4615-8dce-38c2f066d1ac": "Received"},
                    "signalType": "ImagePDQ",
                },
            ],
        }
    return {
        "count": 1,
        "nextSetTimestamp": 1625167824,
        "nextPageToken": PAGE_TOKEN[:-1] + "4",
        "hasMoreRecords": False,
        "hashRecords": [
            {
                "lastModtimestamp": 1625175071,
                "hashValue": "9def0b7dafa86a1c90f2abd78e79ceb25ec3d1a4b3d4bc7a4354baf7717ea038",
                "hashStatus": "Active",
                "caseNumbers": {"cc592711-7068-442f-b5d2-24d50d389751": "Active"},
                "signalType": "ImagePDQ",
                "CSPFeedbacks": [
                    {
                        "source": "Facebook",
                        "feedbackValue": "Blocked",
                        "tags": ["Nude", "Objectionable"],
                    }
                ],
            },
        ],
    }


@pytest.fixture
def api(monkeypatch: pytest.MonkeyPatch):
    api = StopNCIIAPI("", "")
    monkeypatch.setattr(api, "_get", mock_get_impl)
    return api


def assert_first_record(record: StopNCIIHashRecord) -> None:
    assert record.lastModtimestamp == 1625167804
    assert record.signalType == StopNCIISignalType.ImagePDQ
    assert (
        record.hashValue
        == "2afc4a5c09628a7961c14d436493bba66b89b831453baa1d556ba385554daa82"
    )
    assert record.hashStatus == StopNCIICaseStatus.Received
    assert len(record.caseNumbers) == 4
    assert all(c == StopNCIICaseStatus.Received for c in record.caseNumbers.values())
    assert len(record.CSPFeedbacks) == 0


def assert_second_record(record: StopNCIIHashRecord) -> None:
    assert record.lastModtimestamp == 1625167844
    assert record.signalType == StopNCIISignalType.ImagePDQ
    assert (
        record.hashValue
        == "79e07de27d7295339435d63cd31cf35a7bfa29eb2885008500a588a5ea3ae75a"
    )
    assert record.hashStatus == StopNCIICaseStatus.Received
    assert len(record.caseNumbers) == 1
    assert record.caseNumbers == {
        "bd2bff7c-8160-4615-8dce-38c2f066d1ac": StopNCIICaseStatus.Received
    }
    assert len(record.CSPFeedbacks) == 0


def assert_third_record(record: StopNCIIHashRecord) -> None:
    assert record.lastModtimestamp == 1625175071
    assert record.signalType == StopNCIISignalType.ImagePDQ
    assert (
        record.hashValue
        == "9def0b7dafa86a1c90f2abd78e79ceb25ec3d1a4b3d4bc7a4354baf7717ea038"
    )
    assert record.hashStatus == StopNCIICaseStatus.Active
    assert len(record.caseNumbers) == 1
    assert record.caseNumbers == {
        "cc592711-7068-442f-b5d2-24d50d389751": StopNCIICaseStatus.Active
    }

    assert len(record.CSPFeedbacks) == 1
    feedback = record.CSPFeedbacks[0]
    assert feedback.source == "Facebook"
    assert feedback.feedbackValue == StopNCIICSPFeedbackValue.Blocked
    assert feedback.tags == {"Nude", "Objectionable"}


def test_mocked_get_hashes(api: StopNCIIAPI):
    result = api.fetch_hashes()

    assert result.count == 2
    assert result.nextSetTimestamp == 1625175071
    assert result.hasMoreRecords is True
    assert result.nextPageToken == PAGE_TOKEN
    assert len(result.hashRecords) == 2

    one, two = result.hashRecords
    assert_first_record(one)
    assert_second_record(two)

    second_result = api.fetch_hashes(next_page=result.nextPageToken)

    assert second_result.count == 1
    assert second_result.nextSetTimestamp == 1625167824
    assert second_result.hasMoreRecords is False
    assert second_result.nextPageToken != PAGE_TOKEN
    assert len(second_result.hashRecords) == 1
    assert_third_record(second_result.hashRecords[0])


def test_mocked_get_hashes_iter(api: StopNCIIAPI):
    it = api.fetch_hashes_iter()
    as_list = list(it)
    assert len(as_list) == 2

    all_updates = [
        record for result in api.fetch_hashes_iter() for record in result.hashRecords
    ]
    assert len(all_updates) == 3
    one, two, three = all_updates
    assert_first_record(one)
    assert_second_record(two)
    assert_third_record(three)


def mock_get_impl_error_enum(endpoint: str, **json):
    assert endpoint == StopNCIIEndpoint.FetchHashes
    return {
        "count": 3,
        "nextSetTimestamp": 1625167824,
        "nextPageToken": PAGE_TOKEN[:-1] + "4",
        "hasMoreRecords": False,
        "hashRecords": [
            {
                "lastModtimestamp": 1625175071,
                "hashValue": "9def0b7dafa86a1c90f2abd78e79ceb25ec3d1a4b3d4bc7a4354baf7717ea038",
                "hashStatus": "Active",
                "caseNumbers": {"cc592711-7068-442f-b5d2-24d50d389751": "Active"},
                "signalType": "ImagePDQ",
                "CSPFeedbacks": [
                    {
                        "source": "Facebook",
                        "feedbackValue": "Blocked",
                        "tags": ["Nude", "Objectionable"],
                    }
                ],
            },
            {
                "lastModtimestamp": 1625175071,
                "hashValue": "9def0b7dafa86a1c90f2abd78e79ceb25ec3d1a4b3d4bc7a4354baf7717ea038",
                "hashStatus": "Active",
                "caseNumbers": {"cc592711-7068-442f-b5d2-24d50d389751": "Active"},
                "signalType": "ImagePDQ",
                "CSPFeedbacks": [
                    {
                        "source": "Facebook",
                        "feedbackValue": "RandomFeedbackValue",
                        "tags": ["Nude", "Objectionable"],
                    }
                ],
            },
            {
                "lastModtimestamp": 1661426988,
                "hashValue": "95d087d087f8f41cf00771c707f713c7675d07f8c618f00ef00f07f307e37087",
                "hashStatus": "Active",
                "caseNumbers": {
                    "bce008ff-e114-4815-9de8-8f4a50bfd471": "Active",
                    "df018705-fb65-4481-a036-cce0ab02168a": "Active",
                    "504b123a-0fd9-4b95-b7ad-42abf88a1d09": "Received",
                    "ab2754fb-836d-4947-b633-e0c6c7d55255": "Active",
                    "cfaa8342-4f7e-4f5f-88bd-9fd083eafe57": "Active",
                },
                "signalType": "ImagePDQ",
                "hashRegions": ["UK"],
                "CSPFeedbacks": [
                    {
                        "source": "TikTok",
                        "feedbackValue": "Undefined",
                        "tags": ["Tag-One", "Tag-eight"],
                    },
                    {
                        "source": "Twitter",
                        "feedbackValue": "Undefined",
                        "tags": ["Tag-One", "Tag-eight"],
                    },
                ],
            },
        ],
    }


def test_mocked_get_hashes_with_undefined_enum(monkeypatch: pytest.MonkeyPatch):
    error_enum_api = StopNCIIAPI("", "")
    monkeypatch.setattr(error_enum_api, "_get", mock_get_impl_error_enum)
    result = error_enum_api.fetch_hashes()
    assert result.count == 3
    assert len(result.hashRecords) == 1  # since only one record has valid feedbackValue


def mock_feedbacks() -> t.Dict[t.Tuple[StopNCIISignalType, str], StopNCIICSPFeedback]:
    hash_str_1 = "2afc4a5c09628a7961c14d436493bba66b89b831453baa1d556ba385554daa82"
    hash_str_2 = "79e07de27d7295339435d63cd31cf35a7bfa29eb2885008500a588a5ea3ae75a"
    return {
        (StopNCIISignalType.ImagePDQ, hash_str_1): StopNCIICSPFeedback(
            feedbackValue=StopNCIICSPFeedbackValue.Deleted,
            source="facebook",
            tags={"Adult", "Nude"},
        ),
        (StopNCIISignalType.ImagePDQ, hash_str_2): StopNCIICSPFeedback(
            feedbackValue=StopNCIICSPFeedbackValue.Blocked,
            source="snapchat",
            tags={"Nude"},
        ),
    }


def mock_submit_feedback_post_impl(endpoint: str, json):
    # since tags field is `set`, may not inorder, needs to sort in advance
    json["hashFeedbacks"][0]["tags"].sort()
    desired_json = {
        "count": 2,
        "hashFeedbacks": [
            {
                "tags": ["Adult", "Nude"],
                "feedbackValue": "Deleted",
                "hashValue": "2afc4a5c09628a7961c14d436493bba66b89b831453baa1d556ba385554daa82",
            },
            {
                "tags": ["Nude"],
                "feedbackValue": "Blocked",
                "hashValue": "79e07de27d7295339435d63cd31cf35a7bfa29eb2885008500a588a5ea3ae75a",
            },
        ],
    }
    assert endpoint == StopNCIIEndpoint.SubmitFeedback
    assert json == desired_json


@pytest.fixture
def submit_feedback_api(monkeypatch: pytest.MonkeyPatch):
    return submit_feedback_api


def test_post_feedbacks(monkeypatch: pytest.MonkeyPatch):
    submit_feedback_api = StopNCIIAPI("", "")
    monkeypatch.setattr(submit_feedback_api, "_post", mock_submit_feedback_post_impl)
    submit_feedback_api.submit_feedbacks(mock_feedbacks())



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
A solution that allows loading signals directly from a file.

This is useful if you don't have access to an API, but still have a list of 
hashes from somewhere.
"""


import os
import typing as t
from dataclasses import dataclass, field
from pathlib import Path

from threatexchange.exchanges import fetch_state as state
from threatexchange.exchanges.signal_exchange_api import (
    SignalExchangeAPIWithSimpleUpdates,
)
from threatexchange.exchanges.collab_config import CollaborationConfigWithDefaults
from threatexchange.signal_type.signal_base import SignalType

_TypedDelta = state.FetchDelta[
    t.Tuple[str, str],
    state.FetchedSignalMetadata,
    state.FetchCheckpointBase,
]


@dataclass
class _FileCollaborationConfigRequiredFields:
    filename: str = field(
        metadata={"help": "the absolute file path to the signal file"}
    )


@dataclass
class FileCollaborationConfig(
    CollaborationConfigWithDefaults, _FileCollaborationConfigRequiredFields
):
    signal_type: t.Optional[str] = field(
        default=None,
        metadata={
            "help": "if the file row doesn't list the signal type, interpret as this type (SignalType.get_name())",
            "metavar": "SIGNAL_TYPE",
        },
    )


@dataclass
class LocalFileSignalExchangeAPI(
    SignalExchangeAPIWithSimpleUpdates[
        FileCollaborationConfig,
        state.FetchCheckpointBase,
        state.FetchedSignalMetadata,
    ]
):
    """
    Read simple signal files off the local disk.
    """

    collab: FileCollaborationConfig

    @staticmethod
    def get_config_cls() -> t.Type[FileCollaborationConfig]:
        return FileCollaborationConfig

    @staticmethod
    def get_checkpoint_cls() -> t.Type[state.FetchCheckpointBase]:
        return state.FetchCheckpointBase

    @staticmethod
    def get_record_cls() -> t.Type[state.FetchedSignalMetadata]:
        return state.FetchedSignalMetadata

    @classmethod
    def for_collab(
        cls, collab: FileCollaborationConfig
    ) -> "LocalFileSignalExchangeAPI":
        return cls(collab)

    def fetch_iter(
        self,
        _supported_signal_types: t.Sequence[t.Type[SignalType]],
        checkpoint: t.Optional[state.TFetchCheckpoint],
    ) -> t.Iterator[_TypedDelta]:
        """Fetch the whole file"""
        path = Path(self.collab.filename)
        assert path.exists(), f"No such file {path}"
        assert path.is_file(), f"{path} is not a file (is it a dir?)"

        # TODO - Support things other than just one item per line
        with path.open("r") as f:
            lines = f.readlines()

        updates: t.Dict[t.Tuple[str, str], t.Optional[state.FetchedSignalMetadata]] = {}
        for line in lines:
            signal_type = self.collab.signal_type
            signal = line.strip()
            if signal_type is None:
                signal_type, _, signal = signal.partition(" ")
            if signal_type and signal:
                updates[signal_type, signal] = state.FetchedSignalMetadata()

        yield _TypedDelta(updates, state.FetchCheckpointBase())

    def report_opinion(
        self,
        s_type: t.Type[SignalType],
        signal: str,
        opinion: state.SignalOpinion,
    ) -> None:
        if opinion.category != state.SignalOpinionCategory.POSITIVE_CLASS:
            raise NotImplementedError
        if opinion.tags:
            raise NotImplementedError
        path = Path(self.collab.filename)
        with path.open("rb") as rf:
            rf.seek(-1, os.SEEK_END)
            has_newline = rf.read1(1) == b"\n"
        # Appending will overwrite previous ones, and compaction is for scrubs
        with path.open("wta") as wf:
            nl = "" if has_newline else "\n"
            wf.write(f"{nl}{s_type.get_name()} {signal}\n")



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import argparse
import logging
import pathlib
import sys
import tempfile
import shutil
import typing as t

import requests

from threatexchange.cli.exceptions import CommandError

"""
Common helpers for the CLI

Could have been called "common.py"
"""


class FlexFilesInputAction(argparse.Action):
    """
    Takes filenames (or others) and stores path.Path objects.

    An action to support the following argument inputs:

    By default assumes filename
    If `--` is passed in, the remainder is interpreted as text
    If empty, reads from stidin.

    Use with nargs=argparse.REMAINDER for best results, which gives you:

    $ echo "Some text" > file.txt; cmd file.txt
    $ echo "Some text" | cmd
    $ cmd -- Some text

    All producing the same result.
    """

    def __call__(self, _parser, namespace, values, _option_string=None):
        args: t.List[str] = list(values)
        if not args:
            raise argparse.ArgumentError(self, "this argument is required")
        # We have special behavior for -- but argparse sometimes eats it during parsing...
        if "--" in sys.argv and sys.argv[-len(args) - 1] == "--":
            args.insert(0, "--")
        ret = []
        for i, filename in enumerate(args):
            if filename.strip() == "-":
                with tempfile.NamedTemporaryFile("wb", delete=False) as tmp:
                    logging.debug("Writing stdin to %s", tmp.name)
                    shutil.copyfileobj(sys.stdin.buffer, tmp)
                filename = tmp.name
            elif filename.strip() == "--":
                # We could also just open this as a series of streams and seek() them
                with tempfile.NamedTemporaryFile("w", delete=False) as tmp:
                    logging.debug("Writing -- to %s", tmp.name)
                    tmp.write(" ".join(args[i + 1 :]))
                filename = tmp.name
                ret.append(pathlib.Path(tmp.name))
                break
            elif filename.startswith(("http://", "https://")):
                resp = requests.get(filename, allow_redirects=True)
                resp.raise_for_status()
                with tempfile.NamedTemporaryFile("wb", delete=False) as tmp:
                    logging.debug("Writing -- to %s", tmp.name)
                    tmp.write(resp.content)
                filename = pathlib.Path(tmp.name)
            path = pathlib.Path(filename)
            if not path.is_file():
                raise argparse.ArgumentError(self, f"no such file {path}")
            ret.append(path)
        setattr(namespace, self.dest, ret)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the vpdq signal type.
"""

from threatexchange.extensions.vpdq.vpdq_util import (
    VPDQ_INDEX_MATCH_THRESHOLD_PERCENT,
    VpdqCompactFeature,
    json_to_vpdq,
    vpdq_to_json,
    VPDQ_DISTANCE_THRESHOLD,
    VPDQ_QUERY_MATCH_THRESHOLD_PERCENT,
    VPDQ_QUALITY_THRESHOLD,
    hash_file_compact,
)
from threatexchange.extensions.vpdq.vpdq_brute_matcher import match_VPDQ_hash_brute
import pathlib
import typing as t
from threatexchange.content_type.content_base import ContentType
from threatexchange.content_type.video import VideoContent
from threatexchange.signal_type import signal_base
from threatexchange.signal_type.pdq.signal import PdqSignal
from threatexchange.extensions.vpdq.vpdq_index import VPDQSimilarityInfo, VPDQIndex


class VPDQSignal(signal_base.SimpleSignalType, signal_base.FileHasher):
    """
    Simple signal type for video using VPDQ.
    VPDQ signal_str is the json serialization from a list of VPDQ features
    The json serialization will convert the the list of VPDQ features to list of json objects
    where frame_number is the key:
    {(frame_number)
        0: {
            "quality": 100,
            "hash": PDQHash,
            "timestamp": 0.0
            },

        1: {
            "quality": 100,
            "hash": Hash256,
            "timestamp": 0.1
            },
            ...
    }
    Read about VPDQ at https://github.com/facebook/ThreatExchange/tree/main/vpdq
    """

    @classmethod
    def get_content_types(cls) -> t.List[t.Type[ContentType]]:
        return [VideoContent]

    @classmethod
    def get_index_cls(cls):
        return VPDQIndex

    @classmethod
    def validate_signal_str(cls, signal_str: str) -> str:
        """
        @see VpdqCompactFeature
        """
        json_to_vpdq(signal_str)  # throws value error on failure
        return signal_str

    @classmethod
    def hash_from_file(cls, path: pathlib.Path, seconds_per_hash: float = 1) -> str:
        return vpdq_to_json(hash_file_compact(str(path), seconds_per_hash))

    @classmethod
    def compare_hash(
        cls,
        hash1: str,
        hash2: str,
        query_match_pct_thresh: float = VPDQ_QUERY_MATCH_THRESHOLD_PERCENT,
        compare_match_pct_thresh: float = VPDQ_INDEX_MATCH_THRESHOLD_PERCENT,
    ) -> signal_base.SignalComparisonResult:
        vpdq_hash1 = json_to_vpdq(hash1)
        vpdq_hash2 = json_to_vpdq(hash2)
        match_percent = match_VPDQ_hash_brute(
            vpdq_hash1,
            vpdq_hash2,
            VPDQ_QUALITY_THRESHOLD,
            VPDQ_DISTANCE_THRESHOLD,
        )
        return signal_base.SignalComparisonResult(
            (
                match_percent.query_match_percent >= query_match_pct_thresh
                and match_percent.compared_match_percent >= compare_match_pct_thresh
            ),
            VPDQSimilarityInfo(
                match_percent.query_match_percent, match_percent.compared_match_percent
            ),
        )

    @staticmethod
    def get_examples() -> t.List[str]:
        from_pdq_sample = [
            VpdqCompactFeature(pdq_hash, 100, float(timestamp))
            for timestamp, pdq_hash in enumerate(PdqSignal.get_examples())
        ]
        return [vpdq_to_json(from_pdq_sample)]
